# flake8: noqa
"""

DB setup 6.x

Revision ID: 83993fcdf5ef
Revises: cd8b7108713a
Create Date: 2021-10-04 09:22:36.987652

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '83993fcdf5ef'
down_revision = 'cd8b7108713a'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('operations',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('path', sa.String(length=255), nullable=True),
    sa.Column('category', sa.String(length=255), nullable=True),
    sa.Column('description', sa.String(length=255), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('path')
    )
    op.drop_table('projects')
    with op.batch_alter_table('changes', schema=None) as batch_op:
        batch_op.add_column(sa.Column('op_id', sa.Integer(), nullable=True))
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key(None, 'operations', ['op_id'], ['id'])
        batch_op.drop_column('p_id')

    with op.batch_alter_table('messages', schema=None) as batch_op:
        batch_op.add_column(sa.Column('op_id', sa.Integer(), nullable=True))
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key(None, 'operations', ['op_id'], ['id'])
        batch_op.drop_column('p_id')

    with op.batch_alter_table('permissions', schema=None) as batch_op:
        batch_op.add_column(sa.Column('op_id', sa.Integer(), nullable=True))
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key(None, 'operations', ['op_id'], ['id'])
        batch_op.drop_column('p_id')

    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.add_column(sa.Column('registered_on', sa.DateTime(), nullable=False))
        batch_op.add_column(sa.Column('confirmed', sa.Boolean(), nullable=False))
        batch_op.add_column(sa.Column('confirmed_on', sa.DateTime(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.drop_column('confirmed_on')
        batch_op.drop_column('confirmed')
        batch_op.drop_column('registered_on')

    with op.batch_alter_table('permissions', schema=None) as batch_op:
        batch_op.add_column(sa.Column('p_id', sa.INTEGER(), nullable=True))
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key(None, 'projects', ['p_id'], ['id'])
        batch_op.drop_column('op_id')

    with op.batch_alter_table('messages', schema=None) as batch_op:
        batch_op.add_column(sa.Column('p_id', sa.INTEGER(), nullable=True))
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key(None, 'projects', ['p_id'], ['id'])
        batch_op.drop_column('op_id')

    with op.batch_alter_table('changes', schema=None) as batch_op:
        batch_op.add_column(sa.Column('p_id', sa.INTEGER(), nullable=True))
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key(None, 'projects', ['p_id'], ['id'])
        batch_op.drop_column('op_id')

    op.create_table('projects',
    sa.Column('id', sa.INTEGER(), nullable=False),
    sa.Column('path', sa.VARCHAR(length=255), nullable=True),
    sa.Column('description', sa.VARCHAR(length=255), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('path')
    )
    op.drop_table('operations')
    # ### end Alembic commands ###



# flake8: noqa
"""
DB setup until v6.x

Revision ID: cd8b7108713a
Revises: 
Create Date: 2021-10-04 09:10:35.606793

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'cd8b7108713a'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('projects',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('path', sa.String(length=255), nullable=True),
    sa.Column('description', sa.String(length=255), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('path')
    )
    op.create_table('users',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('username', sa.String(length=255), nullable=True),
    sa.Column('emailid', sa.String(length=255), nullable=True),
    sa.Column('password', sa.String(length=255), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('emailid'),
    sa.UniqueConstraint('password')
    )
    op.create_table('changes',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('p_id', sa.Integer(), nullable=True),
    sa.Column('u_id', sa.Integer(), nullable=True),
    sa.Column('commit_hash', sa.String(length=255), nullable=True),
    sa.Column('version_name', sa.String(length=255), nullable=True),
    sa.Column('comment', sa.String(length=255), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.ForeignKeyConstraint(['p_id'], ['projects.id'], ),
    sa.ForeignKeyConstraint(['u_id'], ['users.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('messages',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('p_id', sa.Integer(), nullable=True),
    sa.Column('u_id', sa.Integer(), nullable=True),
    sa.Column('text', sa.Text(), nullable=True),
    sa.Column('message_type', sa.Enum('TEXT', 'SYSTEM_MESSAGE', 'IMAGE', 'DOCUMENT', name='messagetype'), nullable=True),
    sa.Column('reply_id', sa.Integer(), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.ForeignKeyConstraint(['p_id'], ['projects.id'], ),
    sa.ForeignKeyConstraint(['reply_id'], ['messages.id'], ),
    sa.ForeignKeyConstraint(['u_id'], ['users.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('permissions',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('p_id', sa.Integer(), nullable=True),
    sa.Column('u_id', sa.Integer(), nullable=True),
    sa.Column('access_level', sa.Enum('admin', 'collaborator', 'viewer', 'creator', name='access_level'), nullable=True),
    sa.ForeignKeyConstraint(['p_id'], ['projects.id'], ),
    sa.ForeignKeyConstraint(['u_id'], ['users.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('permissions')
    op.drop_table('messages')
    op.drop_table('changes')
    op.drop_table('users')
    op.drop_table('projects')
    # ### end Alembic commands ###



from mslib.utils.units import units
from mslib.mswms.utils import Targets


def test_targets():
    for standard_name in Targets.get_targets():
        unit = Targets.get_unit(standard_name)
        units(unit[0])  # ensure that the unit may be parsed
        assert unit[1] == 1  # no conversion outside pint!
        Targets.get_range(standard_name)
        Targets.get_thresholds(standard_name)
        Targets.get_range(standard_name)
        Targets.UNITS[standard_name]
        Targets.TITLES[standard_name]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    mslib.msui.mss_pyui
    ~~~~~~~~~~~~~~~~~~~

    Mission Support System Python/Qt User Interface
    Main window of the user interface application. Manages view and tool windows
    (the user can open multiple windows) and provides functionality to open, save,
    and switch between flight tracks.

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Hack to fix missing PROJ4 env var in root environment
import os
import sys
import setuptools


if os.getenv("PROJ_LIB") is None or os.getenv("PROJ_LIB") == "PROJ_LIB":
    conda_file_dir = setuptools.__file__
    conda_dir = conda_file_dir.split('lib')[0]
    proj_lib = os.path.join(os.path.join(conda_dir, 'share'), 'proj')
    if "win" in sys.platform:
        proj_lib = os.path.join(os.path.join(conda_dir, 'Library'), 'share')
    os.environ["PROJ_LIB"] = proj_lib
    if not os.path.exists(proj_lib):
        os.makedirs(proj_lib)
        epsg_file = os.path.join(proj_lib, 'epsg')
        if not os.path.exists(epsg_file):
            with open(os.path.join(proj_lib, 'epsg'), 'w') as fid:
                fid.write("# Placeholder for epsg data")


import argparse
import copy
import functools
import hashlib
import importlib
import logging
import os
import platform
import re
import requests
import shutil
import sys
import fs

from mslib import __version__
from mslib.msui.mss_qt import ui_mainwindow as ui
from mslib.msui.mss_qt import ui_about_dialog as ui_ab
from mslib.msui.mss_qt import ui_shortcuts as ui_sh
from mslib.msui import flighttrack as ft
from mslib.msui import tableview, topview, sideview, linearview
from mslib.msui import editor
from mslib.msui import constants
from mslib.msui import wms_control
from mslib.msui import mscolab
from mslib.msui.updater import UpdaterUI
from mslib.utils import setup_logging
from mslib.plugins.io.csv import load_from_csv, save_to_csv
from mslib.msui.icons import icons, python_powered
from mslib.msui.mss_qt import get_open_filename, get_save_filename, Worker, Updater
from mslib.utils.config import read_config_file, config_loader
from PyQt5 import QtGui, QtCore, QtWidgets, QtTest

# Add config path to PYTHONPATH so plugins located there may be found
sys.path.append(constants.MSS_CONFIG_PATH)


def clean_string(string):
    return re.sub(r'\W|^(?=\d)', '_', string)


class QActiveViewsListWidgetItem(QtWidgets.QListWidgetItem):
    """Subclass of QListWidgetItem, represents an open view in the list of
       open views. Keeps a reference to the view instance (i.e. the window) it
       represents in the list of open views.
    """

    # Class variable to assign a unique ID to each view.
    opened_views = 0
    open_views = []

    def __init__(self, view_window, parent=None, viewsChanged=None, mscolab=False,
                 _type=QtWidgets.QListWidgetItem.UserType):
        """Add ID number to the title of the corresponding view window.
        """
        QActiveViewsListWidgetItem.opened_views += 1
        view_name = f"({QActiveViewsListWidgetItem.opened_views:d}) {view_window.name}"
        super(QActiveViewsListWidgetItem, self).__init__(view_name, parent, _type)

        view_window.setWindowTitle(f"({QActiveViewsListWidgetItem.opened_views:d}) {view_window.windowTitle()} - "
                                   f"{view_window.waypoints_model.name}")
        view_window.setIdentifier(view_name)
        self.window = view_window
        self.parent = parent
        self.viewsChanged = viewsChanged
        QActiveViewsListWidgetItem.open_views.append(view_window)

    def view_destroyed(self):
        """Slot that removes this QListWidgetItem from the parent (the
           QListWidget) if the corresponding view has been deleted.
        """
        if self.parent is not None:
            self.parent.takeItem(self.parent.row(self))
            for index, window in enumerate(QActiveViewsListWidgetItem.open_views):
                if window.identifier == self.window.identifier:
                    del QActiveViewsListWidgetItem.open_views[index]
                    break
        if self.viewsChanged is not None:
            self.viewsChanged.emit()


class QFlightTrackListWidgetItem(QtWidgets.QListWidgetItem):
    """Subclass of QListWidgetItem, represents a flight track in the list of
       open flight tracks. Keeps a reference to the flight track instance
       (i.e. the instance of WaypointsTableModel).
    """

    def __init__(self, flighttrack_model, parent=None,
                 type=QtWidgets.QListWidgetItem.UserType):
        """Item class for the list widget that accommodates the open flight
           tracks.

        Arguments:
        flighttrack_model -- instance of a flight track model that is
                             associated with the item
        parent -- pointer to the QListWidgetItem that accommodates this item.
                  If not None, the itemChanged() signal of the parent is
                  connected to the nameChanged() slot of this class, reacting
                  to name changes of the item.
        """
        view_name = flighttrack_model.name
        super(QFlightTrackListWidgetItem, self).__init__(
            view_name, parent, type)

        self.parent = parent
        self.flighttrack_model = flighttrack_model


class MSS_ShortcutsDialog(QtWidgets.QDialog, ui_sh.Ui_ShortcutsDialog):
    """
    Dialog showing shortcuts for all currently open windows
    """
    def __init__(self):
        super(MSS_ShortcutsDialog, self).__init__(QtWidgets.QApplication.activeWindow())
        self.setupUi(self)
        self.current_shortcuts = None
        self.treeWidget.itemDoubleClicked.connect(self.double_clicked)
        self.treeWidget.itemClicked.connect(self.clicked)
        self.leShortcutFilter.textChanged.connect(self.filter_shortcuts)
        self.filterRemoveAction = self.leShortcutFilter.addAction(QtGui.QIcon(icons("64x64", "remove.png")),
                                                                  QtWidgets.QLineEdit.TrailingPosition)
        self.filterRemoveAction.setVisible(False)
        self.filterRemoveAction.setToolTip("Click to remove the filter")
        self.filterRemoveAction.triggered.connect(lambda: self.leShortcutFilter.setText(""))
        self.cbNoShortcut.stateChanged.connect(self.fill_list)
        self.cbAdvanced.stateChanged.connect(lambda i: (self.cbNoShortcut.setVisible(i),
                                                        self.leShortcutFilter.setVisible(i),
                                                        self.cbDisplayType.setVisible(i),
                                                        self.label.setVisible(i),
                                                        self.label_2.setVisible(i),
                                                        self.line.setVisible(i)))
        self.cbDisplayType.currentTextChanged.connect(self.fill_list)
        self.cbAdvanced.stateChanged.emit(self.cbAdvanced.checkState())
        self.oldReject = self.reject
        self.reject = self.custom_reject

    def custom_reject(self):
        """
        Reset highlighted objects when closing the shortcuts dialog
        """
        self.reset_highlight()
        self.oldReject()

    def reset_highlight(self):
        """
        Iterates through all shortcuts and resets the stylesheet
        """
        if self.current_shortcuts:
            for shortcuts in self.current_shortcuts.values():
                for shortcut in shortcuts.values():
                    if shortcut[-1] and hasattr(shortcut[-1], "setStyleSheet"):
                        shortcut[-1].setStyleSheet("")

    def clicked(self, item):
        """
        Highlights the selected item in the GUI as yellow
        """
        self.reset_highlight()
        if hasattr(item, "source_object") and item.source_object and hasattr(item.source_object, "setStyleSheet"):
            item.source_object.setStyleSheet("background-color:yellow;")

    def double_clicked(self, item):
        """
        Executes the shortcut for the doubleclicked item
        """
        if hasattr(item, "source_object") and item.source_object:
            self.reset_highlight()
            self.hide()
            obj = item.source_object
            if isinstance(obj, QtWidgets.QShortcut):
                obj.activated.emit()
            elif isinstance(obj, QtWidgets.QAction):
                obj.trigger()
            elif isinstance(obj, QtWidgets.QAbstractButton):
                obj.click()
            elif isinstance(obj, QtWidgets.QComboBox):
                QtCore.QTimer.singleShot(200, obj.showPopup)
            elif isinstance(obj, QtWidgets.QLineEdit) or isinstance(obj, QtWidgets.QAbstractSpinBox):
                obj.setFocus()

    def fill_list(self):
        """
        Fills the treeWidget with all relevant windows as top level items and their shortcuts as children
        """
        self.reset_highlight()
        self.treeWidget.clear()
        self.current_shortcuts = self.get_shortcuts()
        for widget in self.current_shortcuts:
            if hasattr(widget, "window"):
                name = widget.window().windowTitle()
            else:
                name = widget.objectName()
            if len(name) == 0 or (hasattr(widget, "window") and widget.window().isHidden()):
                continue
            header = QtWidgets.QTreeWidgetItem(self.treeWidget)
            header.setText(0, name)
            if hasattr(widget, "window") and widget.window() == self.parent():
                header.setExpanded(True)
                header.setSelected(True)
                self.treeWidget.setCurrentItem(header)
            for objectName in self.current_shortcuts[widget].keys():
                description, text, _, shortcut, obj = self.current_shortcuts[widget][objectName]
                item = QtWidgets.QTreeWidgetItem(header)
                item.source_object = obj
                itemText = description if self.cbDisplayType.currentText() == 'Tooltip' \
                    else text if self.cbDisplayType.currentText() == 'Text' else objectName
                item.setText(0, f"{itemText}: {shortcut}")
                item.setToolTip(0, f"ToolTip: {description}\nText: {text}\nObjectName: {objectName}")
                header.addChild(item)
        self.filter_shortcuts(self.leShortcutFilter.text())

    def get_shortcuts(self):
        """
        Iterates through all top level widgets and puts their shortcuts in a dictionary
        """
        shortcuts = {}
        for qobject in QtWidgets.QApplication.topLevelWidgets():
            actions = []
            actions.extend([
                (action.parent().window() if hasattr(action.parent(), "window") else action.parent(),
                 action.toolTip(), action.text().replace("&&", "%%").replace("&", "").replace("%%", "&"),
                 action.objectName(),
                 ",".join([shortcut.toString() for shortcut in action.shortcuts()]), action)
                for action in qobject.findChildren(QtWidgets.QAction) if len(action.shortcuts()) > 0 or
                self.cbNoShortcut.checkState()])
            actions.extend([(shortcut.parentWidget().window(), shortcut.whatsThis(), "",
                             shortcut.objectName(), shortcut.key().toString(), shortcut)
                            for shortcut in qobject.findChildren(QtWidgets.QShortcut)])
            actions.extend([(button.window(), button.toolTip(), button.text().replace("&&", "%%").replace("&", "")
                             .replace("%%", "&"), button.objectName(),
                             button.shortcut().toString() if button.shortcut() else "", button)
                            for button in qobject.findChildren(QtWidgets.QAbstractButton) if button.shortcut() or
                            self.cbNoShortcut.checkState()])

            # Additional objects which have no shortcuts, if requested
            actions.extend([(obj.window(), obj.toolTip(), obj.currentText(), obj.objectName(), "", obj)
                            for obj in qobject.findChildren(QtWidgets.QComboBox) if self.cbNoShortcut.checkState()])
            actions.extend([(obj.window(), obj.toolTip(), obj.text(), obj.objectName(), "", obj)
                            for obj in qobject.findChildren(QtWidgets.QAbstractSpinBox) +
                            qobject.findChildren(QtWidgets.QLineEdit)
                            if self.cbNoShortcut.checkState()])
            actions.extend([(obj.window(), obj.toolTip(), obj.toPlainText(), obj.objectName(), "", obj)
                            for obj in qobject.findChildren(QtWidgets.QPlainTextEdit) +
                            qobject.findChildren(QtWidgets.QTextEdit)
                            if self.cbNoShortcut.checkState()])

            if not any(action for action in actions if action[3] == "actionShortcuts"):
                actions.append((qobject.window(), "Show Current Shortcuts", "Show Current Shortcuts",
                               "Show Current Shortcuts", "Alt+S", None))
            if not any(action for action in actions if action[3] == "actionSearch"):
                actions.append((qobject.window(), "Search for interactive text in the UI",
                                "Search for interactive text in the UI", "Search for interactive text in the UI",
                                "Ctrl+F", None))

            for item in actions:
                if item[0] not in shortcuts:
                    shortcuts[item[0]] = {}
                shortcuts[item[0]][item[3].strip()] = item[1:]

        return shortcuts

    def filter_shortcuts(self, text):
        """
        Hides all shortcuts not containing the text
        """
        for window in self.treeWidget.findItems("", QtCore.Qt.MatchContains):
            wms_hits = 0

            for child_index in range(window.childCount()):
                widget = window.child(child_index)
                if text.lower() in widget.text(0).lower() or text.lower() in window.text(0).lower():
                    widget.setHidden(False)
                    wms_hits += 1
                else:
                    widget.setHidden(True)
            if wms_hits == 0 and len(text) > 0:
                window.setHidden(True)
            else:
                window.setHidden(False)

        self.filterRemoveAction.setVisible(len(text) > 0)


class MSS_AboutDialog(QtWidgets.QDialog, ui_ab.Ui_AboutMSUIDialog):
    """Dialog showing information about MSUI. Most of the displayed text is
       defined in the QtDesigner file.
    """

    def __init__(self, parent=None):
        """
        Arguments:
        parent -- Qt widget that is parent to this widget.
        """
        super(MSS_AboutDialog, self).__init__(parent)
        self.setupUi(self)
        self.lblVersion.setText(f"Version: {__version__}")
        self.milestone_url = f'https://github.com/Open-MSS/MSS/issues?q=is%3Aclosed+milestone%3A{__version__[:-1]}'
        self.lblChanges.setText(f'<a href="{self.milestone_url}">New Features and Changes</a>')
        blub = QtGui.QPixmap(python_powered())
        self.lblPython.setPixmap(blub)


class MSSMainWindow(QtWidgets.QMainWindow, ui.Ui_MSSMainWindow):
    """MSUI new main window class. Provides user interface elements for managing
       flight tracks, views and MSColab functionalities.
    """

    viewsChanged = QtCore.pyqtSignal(name="viewsChanged")

    def __init__(self, mscolab_data_dir=None, *args):
        super(MSSMainWindow, self).__init__(*args)
        self.setupUi(self)
        self.setWindowIcon(QtGui.QIcon(icons('32x32')))
        # This code is required in Windows 7 to use the icon set by setWindowIcon in taskbar
        # instead of the default Icon of python/pythonw
        try:
            import ctypes
            myappid = f"mss.mss_pyui.{__version__}"  # arbitrary string
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
        except (ImportError, AttributeError) as error:
            logging.debug("AttributeError, ImportError Exception %s", error)

        self.config_editor = None
        self.local_active = True
        self.new_flight_track_counter = 0

        # Reference to the flight track that is currently displayed in the views.
        self.active_flight_track = None
        self.last_save_directory = config_loader(dataset="data_dir")

        # bind meta (ctrl in macOS) to override automatic translation of ctrl to command by qt
        if sys.platform == 'darwin':
            self.actionTopView.setShortcut(QtGui.QKeySequence("Meta+h"))
            self.actionSideView.setShortcut(QtGui.QKeySequence("Meta+v"))
            self.actionTableView.setShortcut(QtGui.QKeySequence("Meta+t"))
            self.actionLinearView.setShortcut(QtGui.QKeySequence("Meta+l"))
            self.actionConfiguration.setShortcut(QtGui.QKeySequence("Ctrl+,"))

        # File menu.
        self.actionNewFlightTrack.triggered.connect(functools.partial(self.create_new_flight_track, None, None))
        self.actionSaveActiveFlightTrack.triggered.connect(self.save_handler)
        self.actionSaveActiveFlightTrackAs.triggered.connect(self.save_as_handler)
        self.actionCloseSelectedFlightTrack.triggered.connect(self.close_selected_flight_track)

        # Views menu.
        self.actionTopView.triggered.connect(functools.partial(self.create_view_handler, "topview"))
        self.actionSideView.triggered.connect(functools.partial(self.create_view_handler, "sideview"))
        self.actionTableView.triggered.connect(functools.partial(self.create_view_handler, "tableview"))
        self.actionLinearView.triggered.connect(functools.partial(self.create_view_handler, "linearview"))

        # Help menu.
        self.actionOnlineHelp.triggered.connect(self.show_online_help)
        self.actionAboutMSUI.triggered.connect(self.show_about_dialog)
        self.actionShortcuts.triggered.connect(self.show_shortcuts)
        self.actionShortcuts.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionSearch.triggered.connect(lambda: self.show_shortcuts(True))
        self.actionSearch.setShortcutContext(QtCore.Qt.ApplicationShortcut)

        # # Config
        self.actionConfiguration.triggered.connect(self.open_config_editor)

        # Raise Main Window to front with Ctrl/Cmnd + up keyboard shortcut
        self.addAction(self.actionBringMainWindowToFront)
        self.actionBringMainWindowToFront.triggered.connect(self.bring_main_window_to_front)
        self.actionBringMainWindowToFront.setShortcutContext(QtCore.Qt.ApplicationShortcut)

        # Flight Tracks.
        self.listFlightTracks.itemActivated.connect(self.activate_flight_track)

        # Views.
        self.listViews.itemActivated.connect(self.activate_sub_window)

        # Add default and plugins from settings
        picker_default = config_loader(dataset="filepicker_default")
        self.add_plugin_submenu("FTML", "ftml", None, picker_default, plugin_type="Import")
        self.add_plugin_submenu("FTML", "ftml", None, picker_default, plugin_type="Export")
        self.add_plugins()

        preload_urls = config_loader(dataset="WMS_preload")
        self.preload_wms(preload_urls)

        # Status Bar
        self.statusBar.showMessage(self.status())

        # Create MSColab instance to handle all MSColab functionalities
        self.mscolab = mscolab.MSSMscolab(parent=self, data_dir=mscolab_data_dir)

        # Setting up MSColab Tab
        self.connectBtn.clicked.connect(self.mscolab.open_connect_window)

        self.shortcuts_dlg = None

        # deactivate vice versa selection of Operation or Flight Track
        self.listFlightTracks.itemClicked.connect(lambda: self.listOperationsMSC.setCurrentItem(None))
        self.listOperationsMSC.itemClicked.connect(lambda: self.listFlightTracks.setCurrentItem(None))

        # Don't start the updater during a test run of mss_pyui
        if "pytest" not in sys.modules:
            self.updater = UpdaterUI(self)
            self.actionUpdater.triggered.connect(self.updater.show)

    @staticmethod
    def preload_wms(urls):
        """
        This method accesses a list of WMS servers and load their capability documents.
        :param urls: List of URLs
        """
        pdlg = QtWidgets.QProgressDialog("Preloading WMS servers...", "Cancel", 0, len(urls))
        pdlg.reset()
        pdlg.setValue(0)
        pdlg.setModal(True)
        pdlg.show()
        QtWidgets.QApplication.processEvents()
        for i, base_url in enumerate(urls):
            pdlg.setValue(i)
            QtWidgets.QApplication.processEvents()
            # initialize login cache from config file, but do not overwrite existing keys
            for key, value in config_loader(dataset="WMS_login").items():
                if key not in constants.WMS_LOGIN_CACHE:
                    constants.WMS_LOGIN_CACHE[key] = value
            username, password = constants.WMS_LOGIN_CACHE.get(base_url, (None, None))

            try:
                request = requests.get(base_url)
                if pdlg.wasCanceled():
                    break

                wms = wms_control.MSSWebMapService(request.url, version=None,
                                                   username=username, password=password)
                wms_control.WMS_SERVICE_CACHE[wms.url] = wms
                logging.info("Stored WMS info for '%s'", wms.url)
            except Exception as ex:
                logging.error("Error in preloading '%s': '%s'", type(ex), ex)
            if pdlg.wasCanceled():
                break
        logging.debug("Contents of WMS_SERVICE_CACHE: %s", wms_control.WMS_SERVICE_CACHE.keys())
        pdlg.close()

    def bring_main_window_to_front(self):
        self.show()
        self.raise_()
        self.activateWindow()

    def menu_handler(self):
        self.menuImportFlightTrack.setEnabled(True)
        if not self.local_active and self.mscolab.access_level == "viewer":
            # viewer has no import access to server
            self.menuImportFlightTrack.setEnabled(False)

        # enable/disable flight track menus
        self.actionSaveActiveFlightTrack.setEnabled(self.local_active)
        self.actionSaveActiveFlightTrackAs.setEnabled(self.local_active)

    def add_plugins(self):
        picker_default = config_loader(dataset="filepicker_default")
        self.add_plugin_submenu("CSV", "csv", load_from_csv, picker_default, plugin_type="Import")
        self.add_plugin_submenu("CSV", "csv", save_to_csv, picker_default, plugin_type="Export")
        self.import_plugins = {"CSV": (load_from_csv, "csv")}
        self.export_plugins = {"CSV": (save_to_csv, "csv")}
        self.add_import_plugins(picker_default)
        self.add_export_plugins(picker_default)

    def add_plugin_submenu(self, name, extension, function, pickertype, plugin_type="Import"):
        if plugin_type == "Import":
            menu = self.menuImportFlightTrack
            action_name = "actionImportFlightTrack" + clean_string(name)
            handler = self.handle_import_local
        elif plugin_type == "Export":
            menu = self.menuExportActiveFlightTrack
            action_name = "actionExportFlightTrack" + clean_string(name)
            handler = self.handle_export_local

        if hasattr(self, action_name):
            raise ValueError(f"'{action_name}' has already been set!")
        action = QtWidgets.QAction(self)
        action.setObjectName(action_name)
        action.setText(QtCore.QCoreApplication.translate("MSSMainWindow", name, None))
        action.triggered.connect(functools.partial(handler, extension, function, pickertype))
        menu.addAction(action)
        setattr(self, action_name, action)

    def add_import_plugins(self, picker_default):
        plugins = config_loader(dataset="import_plugins")
        for name in plugins:
            extension, module, function = plugins[name][:3]
            picker_type = picker_default
            if len(plugins[name]) == 4:
                picker_type = plugins[name][3]
            try:
                imported_module = importlib.import_module(module)
                imported_function = getattr(imported_module, function)
            # wildcard exception to be resilient against error introduced by user code
            except Exception as ex:
                logging.error("Error on import: %s: %s", type(ex), ex)
                QtWidgets.QMessageBox.critical(
                    self.tr(f"ERROR: Configuration\n\n{plugins,}\n\nthrows {type(ex)} error:\n{ex}"))
                continue
            try:
                self.add_plugin_submenu(name, extension, imported_function, picker_type, plugin_type="Import")
            # wildcard exception to be resilient against error introduced by user code
            except Exception as ex:
                logging.error("Error on installing plugin: %s: %s", type(ex), ex)
                QtWidgets.QMessageBox.critical(
                    self, self.tr("file io plugin error import plugins"),
                    self.tr(f"ERROR: Configuration\n\n{self.import_plugins}\n\nthrows {type(ex)} error:\n{ex}"))
                continue
            self.import_plugins[name] = (imported_function, extension)

    def add_export_plugins(self, picker_default):
        plugins = config_loader(dataset="export_plugins")
        for name in plugins:
            extension, module, function = plugins[name][:3]
            picker_type = picker_default
            if len(plugins[name]) == 4:
                picker_type = plugins[name][3]
            try:
                imported_module = importlib.import_module(module)
                imported_function = getattr(imported_module, function)
            # wildcard exception to be resilient against error introduced by user code
            except Exception as ex:
                logging.error("Error on import: %s: %s", type(ex), ex)
                QtWidgets.QMessageBox.critical(
                    self, self.tr("file io plugin error export plugins"),
                    self.tr(f"ERROR: Configuration\n\n{plugins,}\n\nthrows {type(ex)} error:\n{ex}"))
                continue
            try:
                self.add_plugin_submenu(name, extension, imported_function, picker_type, plugin_type="Export")
            # wildcard exception to be resilient against error introduced by user code
            except Exception as ex:
                logging.error("Error on installing plugin: %s: %s", type(ex), ex)
                QtWidgets.QMessageBox.critical(
                    self, self.tr("file io plugin error import plugins"),
                    self.tr(f"ERROR: Configuration\n\n{self.export_plugins}\n\nthrows {type(ex)} error:\n{ex}"))
                continue
            self.export_plugins[name] = (imported_function, extension)

    def remove_plugins(self):
        for name, _ in self.import_plugins:
            full_name = "actionImportFlightTrack" + clean_string(name)
            actions = [_x for _x in self.menuImportFlightTrack.actions()
                       if _x.objectName() == full_name]
            assert len(actions) == 1
            self.menuImportFlightTrack.removeAction(actions[0])
            delattr(self, full_name)
        self.import_plugins = {}

        for name, _ in self.export_plugins:
            full_name = "actionExportFlightTrack" + clean_string(name)
            actions = [_x for _x in self.menuExportActiveFlightTrack.actions()
                       if _x.objectName() == full_name]
            assert len(actions) == 1
            self.menuExportActiveFlightTrack.removeAction(actions[0])
            delattr(self, full_name)
        self.export_plugins = {}

    def handle_import_local(self, extension, function, pickertype):
        filename = get_open_filename(
            self, "Import Flight Track",
            self.last_save_directory,
            f"Flight Track (*.{extension});;All files (*.*)",
            pickertype=pickertype)
        if self.local_active:
            if filename is not None:
                self.last_save_directory = fs.path.dirname(filename)
                self.create_new_flight_track(filename=filename, function=function)
        else:
            self.mscolab.handle_import_msc(filename, extension, function, pickertype)

    def handle_export_local(self, extension, function, pickertype):
        if self.local_active:
            default_filename = f'{os.path.join(self.last_save_directory, self.active_flight_track.name)}.{extension}'
            filename = get_save_filename(
                self, "Export Flight Track",
                default_filename, f"Flight Track (*.{extension})",
                pickertype=pickertype)
            if filename is not None:
                self.last_save_directory = fs.path.dirname(filename)
                try:
                    if function is None:
                        doc = self.active_flight_track.get_xml_doc()
                        dirname, name = fs.path.split(filename)
                        file_dir = fs.open_fs(dirname)
                        with file_dir.open(name, 'w') as file_object:
                            doc.writexml(file_object, indent="  ", addindent="  ", newl="\n", encoding="utf-8")
                        file_dir.close()
                    else:
                        function(filename, self.active_flight_track.name, self.active_flight_track.waypoints)
                # wildcard exception to be resilient against error introduced by user code
                except Exception as ex:
                    logging.error("file io plugin error: %s %s", type(ex), ex)
                    QtWidgets.QMessageBox.critical(
                        self, self.tr("file io plugin error"),
                        self.tr(f"ERROR: {type(ex)} {ex}"))
        else:
            self.mscolab.handle_export_msc(extension, function, pickertype)

    def create_new_flight_track(self, template=None, filename=None, function=None):
        """Creates a new flight track model from a template. Adds a new entry to
           the list of flight tracks. Called when the user selects the 'new/open
           flight track' menu entries.

        Arguments:
        template -- copy the specified template to the new flight track (so that
                    it is not empty).
        filename -- if not None, load the flight track in the specified file.
        """
        if template is None:
            template = []
            waypoints = config_loader(dataset="new_flighttrack_template")
            default_flightlevel = config_loader(dataset="new_flighttrack_flightlevel")
            for wp in waypoints:
                template.append(ft.Waypoint(flightlevel=default_flightlevel, location=wp))
            if len(template) < 2:
                QtWidgets.QMessageBox.critical(
                    self, self.tr("flighttrack template"),
                    self.tr("ERROR:Flighttrack template in configuration is too short. "
                            "Please add at least two valid locations."))

        waypoints_model = None
        if filename is not None:
            # function is none if ftml file is selected
            if function is None:
                try:
                    waypoints_model = ft.WaypointsTableModel(filename=filename)
                except (SyntaxError, OSError, IOError) as ex:
                    QtWidgets.QMessageBox.critical(
                        self, self.tr("Problem while opening flight track FTML:"),
                        self.tr(f"ERROR: {type(ex)} {ex}"))
            else:
                try:
                    ft_name, new_waypoints = function(filename)
                    waypoints_model = ft.WaypointsTableModel(name=ft_name, waypoints=new_waypoints)
                # wildcard exception to be resilient against error introduced by user code
                except Exception as ex:
                    logging.error("file io plugin error: %s %s", type(ex), ex)
                    QtWidgets.QMessageBox.critical(
                        self, self.tr("file io plugin error"),
                        self.tr(f"ERROR: {type(ex)} {ex}"))
            if waypoints_model is not None:
                for i in range(self.listFlightTracks.count()):
                    fltr = self.listFlightTracks.item(i)
                    if fltr.flighttrack_model.name == waypoints_model.name:
                        waypoints_model.name += " - imported from file"
                        break
        else:
            # Create a new flight track from the waypoints template.
            self.new_flight_track_counter += 1
            waypoints_model = ft.WaypointsTableModel(
                name=f"new flight track ({self.new_flight_track_counter:d})")
            # Make a copy of the template. Otherwise all new flight tracks would
            # use the same data structure in memory.
            template_copy = copy.deepcopy(template)
            waypoints_model.insertRows(0, rows=len(template_copy), waypoints=template_copy)

        if waypoints_model is not None:
            # Create a new list entry for the flight track. Make the item name editable.
            listitem = QFlightTrackListWidgetItem(waypoints_model, self.listFlightTracks)
            listitem.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)

            # Activate new item
            self.activate_flight_track(listitem)

    def activate_flight_track(self, item):
        """Set the currently selected flight track to be the active one, i.e.
           the one that is displayed in the views (only one flight track can be
           displayed at a time).
        """
        self.mscolab.switch_to_local()
        # self.setWindowModality(QtCore.Qt.NonModal)
        self.active_flight_track = item.flighttrack_model
        self.update_active_flight_track()
        font = QtGui.QFont()
        for i in range(self.listFlightTracks.count()):
            self.listFlightTracks.item(i).setFont(font)
        font.setBold(True)
        item.setFont(font)
        self.menu_handler()

    def update_active_flight_track(self, old_flight_track_name=None):
        for i in range(self.listViews.count()):
            view_item = self.listViews.item(i)
            view_item.window.setFlightTrackModel(self.active_flight_track)
            # local we have always all options enabled
            view_item.window.enable_navbar_action_buttons()
            if old_flight_track_name is not None:
                view_item.window.setWindowTitle(view_item.window.windowTitle().replace(old_flight_track_name,
                                                self.active_flight_track.name))

    def activate_selected_flight_track(self):
        item = self.listFlightTracks.currentItem()
        self.activate_flight_track(item)

    def switch_to_mscolab(self):
        self.local_active = False
        font = QtGui.QFont()
        for i in range(self.listFlightTracks.count()):
            self.listFlightTracks.item(i).setFont(font)
        # disable appropriate menu options
        self.menu_handler()

    def save_handler(self):
        """Slot for the 'Save Active Flight Track' menu entry.
        """
        filename = self.active_flight_track.get_filename()
        if filename:
            self.save_flight_track(filename)
        else:
            self.save_as_handler()

    def save_as_handler(self):
        """Slot for the 'Save Active Flight Track As' menu entry.
        """
        default_filename = os.path.join(self.last_save_directory, self.active_flight_track.name + ".ftml")
        file_type = ["Flight track (*.ftml)"]
        filename = get_save_filename(
            self, "Save Flight Track", default_filename, ";;".join(file_type), pickertag="filepicker_default"
        )
        logging.debug("filename : '%s'", filename)
        if filename:
            ext = "ftml"
            self.save_flight_track(filename)
            self.active_flight_track.filename = filename
            self.active_flight_track.name = fs.path.basename(filename.replace(f"{ext}", "").strip())

    def save_flight_track(self, file_name):
        ext = ".ftml"
        if file_name:
            if file_name.endswith(ext):
                try:
                    self.active_flight_track.save_to_ftml(file_name)
                except (OSError, IOError) as ex:
                    QtWidgets.QMessageBox.critical(
                        self, self.tr("Problem while saving flight track to FTML:"),
                        self.tr(f"ERROR: {type(ex)} {ex}"))

            for idx in range(self.listFlightTracks.count()):
                if self.listFlightTracks.item(idx).flighttrack_model == self.active_flight_track:
                    old_filght_track_name = self.listFlightTracks.item(idx).text()
                    self.listFlightTracks.item(idx).setText(self.active_flight_track.name)

            self.update_active_flight_track(old_filght_track_name)

    def close_selected_flight_track(self):
        """Slot to close the currently selected flight track. Flight tracks can
           only be closed if at least one other flight track remains open. The
           currently active flight track cannot be closed.
        """
        if self.listFlightTracks.count() < 2:
            QtWidgets.QMessageBox.information(self, self.tr("Flight Track Management"),
                                              self.tr("At least one flight track has to be open."))
            return
        item = self.listFlightTracks.currentItem()
        if item.flighttrack_model == self.active_flight_track and self.local_active:
            QtWidgets.QMessageBox.information(self, self.tr("Flight Track Management"),
                                              self.tr("Cannot close currently active flight track."))
            return
        if item.flighttrack_model.modified:
            ret = QtWidgets.QMessageBox.warning(self, self.tr("Mission Support System"),
                                                self.tr("The flight track you are about to close has "
                                                        "been modified. Close anyway?"),
                                                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                                                QtWidgets.QMessageBox.No)
            if ret == QtWidgets.QMessageBox.Yes:
                self.listFlightTracks.takeItem(self.listFlightTracks.currentRow())

    def create_view_handler(self, _type):
        if self.local_active:
            self.create_view(_type, self.active_flight_track)
        else:
            self.mscolab.waypoints_model.name = self.mscolab.active_operation_name
            self.create_view(_type, self.mscolab.waypoints_model)

    def create_view(self, _type, model):
        """Method called when the user selects a new view to be opened. Creates
           a new instance of the view and adds a QActiveViewsListWidgetItem to
           the list of open views (self.listViews).
        """
        layout = config_loader(dataset="layout")
        view_window = None
        if _type == "topview":
            # Top view.
            view_window = topview.MSSTopViewWindow(model=model)
            view_window.mpl.resize(layout['topview'][0], layout['topview'][1])
            if layout["immutable"]:
                view_window.mpl.setFixedSize(layout['topview'][0], layout['topview'][1])
        elif _type == "sideview":
            # Side view.
            view_window = sideview.MSSSideViewWindow(model=model)
            view_window.mpl.resize(layout['sideview'][0], layout['sideview'][1])
            if layout["immutable"]:
                view_window.mpl.setFixedSize(layout['sideview'][0], layout['sideview'][1])
        elif _type == "tableview":
            # Table view.
            view_window = tableview.MSSTableViewWindow(model=model)
            view_window.centralwidget.resize(layout['tableview'][0], layout['tableview'][1])
        elif _type == "linearview":
            # Linear view.
            view_window = linearview.MSSLinearViewWindow(model=model)
            view_window.mpl.resize(layout['linearview'][0], layout['linearview'][1])
            if layout["immutable"]:
                view_window.mpl.setFixedSize(layout['linearview'][0], layout['linearview'][1])

        if view_window is not None:
            # Set view type to window
            view_window.view_type = view_window.name
            # Make sure view window will be deleted after being closed, not
            # just hidden (cf. Chapter 5 in PyQt4).
            view_window.setAttribute(QtCore.Qt.WA_DeleteOnClose)
            # Open as a non-modal window.
            view_window.show()
            # Add an entry referencing the new view to the list of views.
            # listitem = QActiveViewsListWidgetItem(view_window, self.listViews, self.viewsChanged, mscolab)
            listitem = QActiveViewsListWidgetItem(view_window, self.listViews, self.viewsChanged)
            view_window.viewCloses.connect(listitem.view_destroyed)
            self.listViews.setCurrentItem(listitem)
            # self.active_view_windows.append(view_window)
            # disable navbar actions in the view for viewer
            try:
                if self.mscolab.access_level == "viewer":
                    view_window.disable_navbar_action_buttons()
            except AttributeError:
                view_window.enable_navbar_action_buttons()
            self.viewsChanged.emit()

    def get_active_views(self):
        active_view_windows = []
        for i in range(self.listViews.count()):
            active_view_windows.append(self.listViews.item(i).window)
        return active_view_windows

    def activate_sub_window(self, item):
        """When the user clicks on one of the open view or tool windows, this
           window is brought to the front. This function implements the slot to
           activate a window if the user selects it in the list of views or
           tools.
        """
        # Restore the activated view and bring it to the front.
        item.window.showNormal()
        item.window.raise_()
        item.window.activateWindow()

    def restart_application(self):
        while self.listViews.count() > 0:
            self.listViews.item(0).window.handle_force_close()
        self.listViews.clear()
        self.remove_plugins()
        self.add_plugins()
        if self.mscolab.token is not None:
            self.mscolab.logout()
        read_config_file()

    def open_config_editor(self):
        """
        Opens up a JSON config editor
        """
        if self.config_editor is None:
            self.config_editor = editor.ConfigurationEditorWindow(parent=self)
            self.config_editor.setAttribute(QtCore.Qt.WA_DeleteOnClose)
            self.config_editor.destroyed.connect(self.close_config_editor)
            self.config_editor.restartApplication.connect(self.restart_application)
            self.config_editor.show()
        else:
            self.config_editor.showNormal()
            self.config_editor.activateWindow()

    def close_config_editor(self):
        self.config_editor = None

    def show_online_help(self):
        """Open Documentation in a browser"""
        QtGui.QDesktopServices.openUrl(
            QtCore.QUrl("http://mss.readthedocs.io/en/stable"))

    def show_about_dialog(self):
        """Show the 'About MSUI' dialog to the user.
        """
        dlg = MSS_AboutDialog(parent=self)
        dlg.setModal(True)
        dlg.exec_()

    def show_shortcuts(self, search_mode=False):
        """Show the shortcuts dialog to the user.
        """
        if QtWidgets.QApplication.activeWindow() == self.shortcuts_dlg:
            return

        self.shortcuts_dlg = MSS_ShortcutsDialog() if not self.shortcuts_dlg else self.shortcuts_dlg

        # In case the dialog gets deleted by QT, recreate it
        try:
            self.shortcuts_dlg.setModal(True)
        except RuntimeError:
            self.shortcuts_dlg = MSS_ShortcutsDialog()

        self.shortcuts_dlg.setParent(QtWidgets.QApplication.activeWindow(), QtCore.Qt.Dialog)
        self.shortcuts_dlg.fill_list()
        self.shortcuts_dlg.show()
        if search_mode:
            self.shortcuts_dlg.cbDisplayType.setCurrentIndex(1)
            self.shortcuts_dlg.leShortcutFilter.setText("")
            self.shortcuts_dlg.cbAdvanced.setCheckState(2)
            self.shortcuts_dlg.cbNoShortcut.setCheckState(2)
            self.shortcuts_dlg.leShortcutFilter.setFocus()

    def status(self):
        if config_loader() != config_loader(default=True):
            return ("Status : System Configuration")
        else:
            return (f"Status : User Configuration '{constants.MSS_SETTINGS}' loaded")

    def closeEvent(self, event):
        """Ask user if he/she wants to close the application. If yes, also
           close all views that are open.

        Overloads QtGui.QMainWindow.closeEvent(). This method is called if
        Qt receives a window close request for our application window.
        """
        ret = QtWidgets.QMessageBox.warning(
            self, self.tr("Mission Support System"),
            self.tr("Do you want to close the Mission Support System application?"),
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No, QtWidgets.QMessageBox.No)

        if ret == QtWidgets.QMessageBox.Yes:
            if self.mscolab.help_dialog is not None:
                self.mscolab.help_dialog.close()
            # cleanup mscolab widgets
            if self.mscolab.token is not None:
                self.mscolab.logout()
            # Table View stick around after MainWindow closes - maybe some dangling reference?
            # This removes them for sure!
            while self.listViews.count() > 0:
                self.listViews.item(0).window.handle_force_close()
            self.listViews.clear()
            self.listFlightTracks.clear()
            # close configuration editor
            if self.config_editor is not None:
                self.config_editor.restart_on_save = False
                self.config_editor.close()
                QtTest.QTest.qWait(5)
                if self.config_editor is not None:
                    self.statusBar.showMessage("Save your config changes and try closing again")
                    event.ignore()
                    return
            event.accept()
        else:
            event.ignore()


def main():
    try:
        prefix = os.environ["CONDA_DEFAULT_ENV"]
    except KeyError:
        prefix = ""
    app_prefix = prefix
    if prefix:
        app_prefix = f"-{prefix}"
    icon_hash = hashlib.md5('.'.join([__version__, app_prefix]).encode('utf-8')).hexdigest()

    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--version", help="show version", action="store_true", default=False)
    parser.add_argument("--debug", help="show debugging log messages on console", action="store_true", default=False)
    parser.add_argument("--logfile", help="Specify logfile location. Set to empty string to disable.", action="store",
                        default=os.path.join(constants.MSS_CONFIG_PATH, "mss_pyui.log"))
    parser.add_argument("-m", "--menu", help="adds mss to menu", action="store_true", default=False)
    parser.add_argument("-d", "--deinstall", help="removes mss from menu", action="store_true", default=False)
    parser.add_argument("--update", help="Updates MSS to the newest version", action="store_true", default=False)

    args = parser.parse_args()

    if args.version:
        print("***********************************************************************")
        print("\n            Mission Support System (mss)\n")
        print("***********************************************************************")
        print("Documentation: http://mss.rtfd.io")
        print("Version:", __version__)
        sys.exit()

    if args.update:
        updater = Updater()
        updater.on_update_available.connect(lambda old, new: updater.update_mss())
        updater.on_log_update.connect(lambda s: print(s.replace("\n", "")))
        updater.on_status_update.connect(lambda s: print(s.replace("\n", "")))
        updater.run()
        while Worker.workers:
            list(Worker.workers)[0].wait()
        sys.exit()

    setup_logging(args)

    if args.menu:
        # Experimental feature to get mss into application menu
        if platform.system() == "Linux":
            icon_size = '48x48'
            src_icon_path = icons(icon_size)
            icon_destination = constants.POSIX["icon_destination"].format(icon_size, icon_hash)
            dirname = os.path.dirname(icon_destination)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            shutil.copyfile(src_icon_path, icon_destination)
            desktop = constants.POSIX["desktop"]
            application_destination = constants.POSIX["application_destination"].format(app_prefix)
            dirname = os.path.dirname(application_destination)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            if prefix:
                prefix = f"({prefix})"
            desktop = desktop.format(prefix,
                                     os.path.join(sys.prefix, "bin", "mss"),
                                     icon_destination)
            with open(application_destination, 'w') as f:
                f.write(desktop)
            logging.info("menu entry created")
        sys.exit()
    if args.deinstall:
        application_destination = constants.POSIX["application_destination"].format(app_prefix)
        if os.path.exists(application_destination):
            os.remove(application_destination)
        icon_size = '48x48'
        icon_destination = constants.POSIX["icon_destination"].format(icon_size, icon_hash)
        if os.path.exists(icon_destination):
            os.remove(icon_destination)
        logging.info("menu entry removed")
        sys.exit()

    logging.info("MSS Version: %s", __version__)
    logging.info("Python Version: %s", sys.version)
    logging.info("Platform: %s (%s)", platform.platform(), platform.architecture())

    try:
        read_config_file()
    except (FileNotFoundError, fs.errors.CreateFailed, fs.errors.FileExpected) as ex:
        message = f'\n\nFix the setup of your "MSS_SETTINGS" configuration.\n{ex}'
        logging.error(message)
        sys.exit()

    application = QtWidgets.QApplication(sys.argv)
    mainwindow = None

    # Trigger shortcuts/search dialog even on modal dialogs
    application.oldNotify = application.notify

    def notify(QObject, QEvent):
        if QEvent.type() == QtCore.QEvent.KeyPress and mainwindow:
            if QEvent.key() == QtCore.Qt.Key_S and QEvent.modifiers() == QtCore.Qt.AltModifier:
                mainwindow.show_shortcuts()
            elif QEvent.key() == QtCore.Qt.Key_F and QEvent.modifiers() == QtCore.Qt.ControlModifier:
                mainwindow.show_shortcuts(True)
        return application.oldNotify(QObject, QEvent)

    application.notify = notify

    application.setWindowIcon(QtGui.QIcon(icons('128x128')))
    application.setApplicationDisplayName("MSS")
    application.setAttribute(QtCore.Qt.AA_DisableWindowContextHelpButton)
    mainwindow = MSSMainWindow()
    mainwindow.setStyleSheet("QListWidget { border: 1px solid grey; }")
    mainwindow.create_new_flight_track()
    mainwindow.show()
    sys.exit(application.exec_())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""

    mslib.mscolab.chat_manager
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Code to handle socket connections in mscolab

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import datetime

import fs

from mslib.mscolab.conf import mscolab_settings
from mslib.mscolab.models import db, Message, MessageType
from mslib.mscolab.utils import get_message_dict


class ChatManager(object):
    """Class with handler functions for chat related functionalities"""

    def __init__(self):
        pass

    def add_message(self, user, text, operation_name, message_type=MessageType.TEXT, reply_id=None):
        """
        text: message to be emitted to operation and saved to db
        operation_name: operation-name(op_id) to which message is emitted,
        user: User object, one which emits the message
        message_type: Enum of type MessageType. values: TEXT, SYSTEM_MESSAGE, IMAGE, DOCUMENT
        """
        if reply_id == -1:
            reply_id = None
        message = Message(operation_name, user.id, text, message_type, reply_id)
        db.session.add(message)
        db.session.commit()
        return message

    def get_messages(self, op_id, timestamp=None):
        """
        op_id: operation id
        timestamp:  if provided, messages only after this time stamp is provided
        """
        if timestamp is None:
            timestamp = datetime.datetime(1970, 1, 1)
        else:
            timestamp = datetime.datetime.strptime(timestamp, "%Y-%m-%d, %H:%M:%S")
        messages = Message.query \
            .filter(Message.op_id == op_id) \
            .filter(Message.reply_id.is_(None)) \
            .filter(Message.created_at > timestamp) \
            .all()

        message_list = []
        for message in messages:
            replies_list = []
            for reply in message.replies:
                reply_dict = get_message_dict(reply)
                replies_list.append(reply_dict)
            message_dict = get_message_dict(message)
            message_dict["replies"] = replies_list
            message_list.append(message_dict)

        return message_list

    def edit_message(self, message_id, new_message_text):
        message = Message.query.filter_by(id=message_id).first()
        message.text = new_message_text
        db.session.commit()

    def delete_message(self, message_id):
        message = Message.query.filter(Message.id == message_id).first()
        if message.message_type == MessageType.IMAGE or message.message_type == MessageType.DOCUMENT:
            file_name = fs.path.basename(message.text)
            with fs.open_fs(mscolab_settings.UPLOAD_FOLDER) as upload_dir:
                upload_dir.remove(fs.path.join(str(message.op_id), file_name))
        db.session.delete(message)
        db.session.commit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

    mslib.mswms.mswms
    ~~~~~~~~~~~~~~~~~

    The module can be run with the Python Flask framework and can be run as
    python mswms.py.

    :copyright: Copyright 2016 Reimar Bauer
    :copyright: Copyright 2016-2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import argparse
import logging
import sys

from mslib import __version__
from mslib.mswms.wms import mss_wms_settings, server
from mslib.mswms.wms import app as application
from mslib.utils import setup_logging
from mslib.msui.mss_qt import Updater, Worker


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--version", help="show version", action="store_true", default=False)
    parser.add_argument("--host", help="hostname",
                        default="127.0.0.1", dest="host")
    parser.add_argument("--port", help="port", dest="port", default="8081")
    parser.add_argument("--threadpool", help="threadpool", dest="use_threadpool", action="store_true", default=False)
    parser.add_argument("--debug", help="show debugging log messages on console", action="store_true", default=False)
    parser.add_argument("--logfile", help="If set to a name log output goes to that file", dest="logfile",
                        default=None)
    parser.add_argument("--update", help="Updates MSS to the newest version", action="store_true", default=False)

    subparsers = parser.add_subparsers(help='Available actions', dest='action')
    gallery = subparsers.add_parser("gallery", help="Subcommands surrounding the gallery")
    gallery.add_argument("--create", action="store_true", default=False,
                         help="Generates plots of all layers not already present")
    gallery.add_argument("--clear", action="store_true", default=False,
                         help="Deletes all plots and corresponding code")
    gallery.add_argument("--refresh", action="store_true", default=False,
                         help="Deletes all plots and regenerates them, a mix of --clear and --create")
    gallery.add_argument("--levels", default="", help="A comma-separated list of all levels visible on the gallery.\n"
                                                      "E.g. --levels 200,300"
                                                      "Use --levels all to include all levels.\n"
                                                      "Default is the middle level.")
    gallery.add_argument("--itimes", default="", help="A comma-separated list of all init times visible on the gallery"
                                                      ", in ISO format.\nE.g. --itimes 2012-10-17T12:00:00\n"
                                                      "Use --itimes all to use all available itimes.\n"
                                                      "Default is the latest itime.")
    gallery.add_argument("--vtimes", default="", help="A comma-separated list of all valid times visible on the gallery"
                                                      ", in ISO format.\nE.g. --vtimes 2012-10-19T12:00:00\n"
                                                      "Use --vtimes all to use all available vtimes.\n"
                                                      "Default is the latest vtime")
    gallery.add_argument("--show-code", action="store_true", default=False,
                         help="Generates plots of all layers not already present, "
                              "and generates code snippets for each plot when clicking on the image")
    gallery.add_argument("--url-prefix", default="",
                         help="Normally the plot images should appear at the relative url /static/plots/*.png.\n"
                              "In case they are prefixed by something, e.g. /demo/static/plots/*.png,"
                              " please provide the prefix /demo here.")

    args = parser.parse_args()

    if args.version:
        print("***********************************************************************")
        print("\n            Mission Support System (mss)\n")
        print("***********************************************************************")
        print("Documentation: http://mss.rtfd.io")
        print("Version:", __version__)
        sys.exit()

    updater = Updater()
    if args.update:
        updater.on_update_available.connect(lambda old, new: updater.update_mss())
        updater.on_log_update.connect(lambda s: print(s.replace("\n", "")))
        updater.on_status_update.connect(lambda s: print(s.replace("\n", "")))
        updater.run()
        while Worker.workers:
            list(Worker.workers)[0].wait()
        sys.exit()

    setup_logging(args)

    if args.action == "gallery":
        create = args.create or args.refresh
        clear = args.clear or args.refresh
        server.generate_gallery(create, clear, args.show_code, url_prefix=args.url_prefix, levels=args.levels,
                                itimes=args.itimes, vtimes=args.vtimes)
        logging.info("Gallery generation done.")
        sys.exit()

    updater.on_update_available.connect(lambda old, new: logging.info(f"MSS can be updated from {old} to {new}.\nRun"
                                                                      " the --update argument to update the server."))
    updater.run()

    logging.info("Configuration File: '%s'", mss_wms_settings.__file__)

    application.run(args.host, args.port)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""

    mslib.msui._tests.test_mss_pyui
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module provides pytest functions to tests msui.mss_pyui

    This file is part of mss.

    :copyright: Copyright 2017 Joern Ungermann
    :copyright: Copyright 2017-2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""


import sys
import mock
import os
import pytest
from urllib.request import urlopen
from PyQt5 import QtWidgets, QtTest
from mslib import __version__
from mslib._tests.constants import ROOT_DIR
import mslib.msui.mss_pyui as mss_pyui
from mslib._tests.utils import ExceptionMock


class Test_MSS_AboutDialog():
    def setup(self):
        self.application = QtWidgets.QApplication(sys.argv)
        self.window = mss_pyui.MSS_AboutDialog()

    def test_milestone_url(self):
        with urlopen(self.window.milestone_url) as f:
            text = f.read()
        pattern = f'value="is:closed milestone:{__version__[:-1]}"'
        assert pattern in text.decode('utf-8')

    def teardown(self):
        self.window.hide()
        QtWidgets.QApplication.processEvents()
        self.application.quit()
        QtWidgets.QApplication.processEvents()


class Test_MSS_ShortcutDialog():
    def setup(self):
        self.application = QtWidgets.QApplication(sys.argv)
        self.main_window = mss_pyui.MSSMainWindow()
        self.main_window.show()
        self.shortcuts = mss_pyui.MSS_ShortcutsDialog()

    def teardown(self):
        self.shortcuts.hide()
        self.main_window.hide()
        QtWidgets.QApplication.processEvents()
        self.application.quit()
        QtWidgets.QApplication.processEvents()

    def test_shortcuts_present(self):
        # Assert list gets filled properly
        self.shortcuts.fill_list()
        assert self.shortcuts.treeWidget.topLevelItemCount() == 1
        self.shortcuts.filter_shortcuts("Nothing")
        assert self.shortcuts.treeWidget.topLevelItem(0).isHidden()

        # Assert changing display type works
        self.shortcuts.cbAdvanced.click()
        old_text = self.shortcuts.treeWidget.topLevelItem(0).child(1).text(0)
        self.shortcuts.cbDisplayType.setCurrentIndex(2)
        assert self.shortcuts.treeWidget.topLevelItem(0).child(1).text(0) != old_text

        # Assert double clicking works
        self.shortcuts.cbNoShortcut.setCheckState(True)
        self.shortcuts.leShortcutFilter.setText("actionConfiguration")
        for i in range(self.shortcuts.treeWidget.topLevelItem(0).childCount()):
            child = self.shortcuts.treeWidget.topLevelItem(0).child(i)
            if not child.isHidden():
                self.shortcuts.double_clicked(child)
                self.shortcuts.fill_list()
                break
        assert self.shortcuts.treeWidget.topLevelItemCount() == 2


class Test_MSSSideViewWindow(object):
    # temporary file paths to test open feature
    sample_path = os.path.join(os.path.dirname(__file__), "..", "..", "..", "docs", "samples", "flight-tracks")
    open_csv = os.path.join(sample_path, "example.csv")
    open_ftml = os.path.join(sample_path, "example.ftml")
    open_txt = os.path.join(sample_path, "example.txt")
    open_fls = os.path.join(sample_path, "flitestar.txt")
    # temporary file paths to test save feature
    save_csv = os.path.join(ROOT_DIR, "example.csv")
    save_ftml = os.path.join(ROOT_DIR, "example.ftml")
    save_ftml = save_ftml.replace('\\', '/')
    save_txt = os.path.join(ROOT_DIR, "example.txt")
    # import/export plugins
    import_plugins = {
        "Text": ["txt", "mslib.plugins.io.text", "load_from_txt"],
        "FliteStar": ["txt", "mslib.plugins.io.flitestar", "load_from_flitestar"],
    }
    export_plugins = {
        "Text": ["txt", "mslib.plugins.io.text", "save_to_txt"],
        # "KML": ["kml", "mslib.plugins.io.kml", "save_to_kml"],
        # "GPX": ["gpx", "mslib.plugins.io.gpx", "save_to_gpx"]
    }

    def setup(self):
        self.application = QtWidgets.QApplication(sys.argv)

        self.window = mss_pyui.MSSMainWindow()
        self.window.create_new_flight_track()
        self.window.show()
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWaitForWindowExposed(self.window)
        QtWidgets.QApplication.processEvents()

    def teardown(self):
        for i in range(self.window.listViews.count()):
            self.window.listViews.item(i).window.hide()
        self.window.hide()
        QtWidgets.QApplication.processEvents()
        self.application.quit()
        QtWidgets.QApplication.processEvents()

    def test_no_updater(self):
        assert not hasattr(self.window, "updater")

    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    def test_app_start(self, mockbox):
        assert mockbox.critical.call_count == 0

    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    def test_new_flightrack(self, mockbox):
        assert self.window.listFlightTracks.count() == 1
        self.window.actionNewFlightTrack.trigger()
        QtWidgets.QApplication.processEvents()
        assert self.window.listFlightTracks.count() == 2
        assert mockbox.critical.call_count == 0

    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    def test_open_topview(self, mockbox):
        assert self.window.listViews.count() == 0
        self.window.actionTopView.trigger()
        QtWidgets.QApplication.processEvents()
        assert mockbox.critical.call_count == 0
        assert self.window.listViews.count() == 1

    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    def test_open_sideview(self, mockbox):
        assert self.window.listViews.count() == 0
        self.window.actionSideView.trigger()
        QtWidgets.QApplication.processEvents()
        assert mockbox.critical.call_count == 0
        assert self.window.listViews.count() == 1

    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    def test_open_tableview(self, mockbox):
        assert self.window.listViews.count() == 0
        self.window.actionTableView.trigger()
        QtWidgets.QApplication.processEvents()
        assert mockbox.critical.call_count == 0
        assert self.window.listViews.count() == 1

    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    def test_open_linearview(self, mockbox):
        assert self.window.listViews.count() == 0
        self.window.actionLinearView.trigger()
        self.window.listViews.itemActivated.emit(self.window.listViews.item(0))
        QtWidgets.QApplication.processEvents()
        assert self.window.listViews.count() == 1
        assert mockbox.critical.call_count == 0

    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    def test_open_about(self, mockbox):
        self.window.actionAboutMSUI.trigger()
        QtWidgets.QApplication.processEvents()
        assert mockbox.critical.call_count == 0

    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    def test_open_config(self, mockbox):
        pytest.skip("To be done")
        self.window.actionConfigurationEditor.trigger()
        QtWidgets.QApplication.processEvents()
        self.window.config_editor.close()
        assert mockbox.critical.call_count == 0

    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    def test_open_shortcut(self, mockbox):
        self.window.actionShortcuts.trigger()
        QtWidgets.QApplication.processEvents()
        assert mockbox.critical.call_count == 0

    @pytest.mark.parametrize("save_file", [[save_ftml]])
    def test_plugin_saveas(self, save_file):
        with mock.patch("mslib.msui.mss_pyui.config_loader", return_value=self.export_plugins):
            self.window.add_export_plugins("qt")
        with mock.patch("mslib.msui.mss_pyui.get_save_filename", return_value=save_file[0]) as mocksave:
            assert self.window.listFlightTracks.count() == 1
            assert mocksave.call_count == 0
            self.window.last_save_directory = ROOT_DIR
            self.window.actionSaveActiveFlightTrackAs.trigger()
            QtWidgets.QApplication.processEvents()
            assert mocksave.call_count == 1
            assert os.path.exists(save_file[0])
            os.remove(save_file[0])

    @pytest.mark.parametrize(
        "open_file", [(open_ftml, "actionImportFlightTrackFTML"), (open_csv, "actionImportFlightTrackCSV"),
                      (open_txt, "actionImportFlightTrackText"), (open_fls, "actionImportFlightTrackFliteStar")])
    def test_plugin_import(self, open_file):
        with mock.patch("mslib.msui.mss_pyui.config_loader", return_value=self.import_plugins):
            self.window.add_import_plugins("qt")
        with mock.patch("mslib.msui.mss_pyui.get_open_filename", return_value=open_file[0]) as mockopen:
            assert self.window.listFlightTracks.count() == 1
            assert mockopen.call_count == 0
            self.window.last_save_directory = ROOT_DIR
            obj_name = open_file[1]
            for action in self.window.menuImportFlightTrack.actions():
                if obj_name == action.objectName():
                    action.trigger()
                    break
            QtWidgets.QApplication.processEvents()
            assert mockopen.call_count == 1
            assert self.window.listFlightTracks.count() == 2

    @pytest.mark.parametrize("save_file", [[save_ftml, "actionExportFlightTrackFTML"],
                                           [save_csv, "actionExportFlightTrackCSV"],
                                           [save_txt, "actionExportFlightTrackText"]])
    def test_plugin_export(self, save_file):
        with mock.patch("mslib.msui.mss_pyui.config_loader", return_value=self.export_plugins):
            self.window.add_export_plugins("qt")
        with mock.patch("mslib.msui.mss_pyui.get_save_filename", return_value=save_file[0]) as mocksave:
            assert self.window.listFlightTracks.count() == 1
            assert mocksave.call_count == 0
            self.window.last_save_directory = ROOT_DIR
            obj_name = save_file[1]
            for action in self.window.menuExportActiveFlightTrack.actions():
                if obj_name == action.objectName():
                    action.trigger()
                    break
            QtWidgets.QApplication.processEvents()
            assert mocksave.call_count == 1
            assert os.path.exists(save_file[0])
            os.remove(save_file[0])

    @pytest.mark.skip("needs to be refactored to become independent")
    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    @mock.patch("mslib.msui.mss_pyui.config_loader", return_value=export_plugins)
    def test_add_plugins(self, mockopen, mockbox):
        assert len(self.window.menuImportFlightTrack.actions()) == 2
        assert len(self.window.menuExportActiveFlightTrack.actions()) == 2
        assert len(self.window.import_plugins) == 1
        assert len(self.window.export_plugins) == 1

        self.window.remove_plugins()
        self.window.add_import_plugins("qt")
        self.window.add_export_plugins("qt")
        assert len(self.window.import_plugins) == 1
        assert len(self.window.export_plugins) == 1
        assert len(self.window.menuImportFlightTrack.actions()) == 2
        assert len(self.window.menuExportActiveFlightTrack.actions()) == 2
        assert mockbox.critical.call_count == 0

        self.window.remove_plugins()
        with mock.patch("importlib.import_module", new=ExceptionMock(Exception()).raise_exc):
            self.window.add_import_plugins("qt")
            self.window.add_export_plugins("qt")
            assert mockbox.critical.call_count == 2

        self.window.remove_plugins()
        with mock.patch("mslib.msui.mss_pyui.MSSMainWindow.add_plugin_submenu",
                        new=ExceptionMock(Exception()).raise_exc):
            self.window.add_import_plugins("qt")
            self.window.add_export_plugins("qt")
            assert mockbox.critical.call_count == 4

        self.window.remove_plugins()
        assert len(self.window.import_plugins) == 0
        assert len(self.window.export_plugins) == 0
        assert len(self.window.menuImportFlightTrack.actions()) == 1
        assert len(self.window.menuExportActiveFlightTrack.actions()) == 1

    @mock.patch("PyQt5.QtWidgets.QMessageBox.critical")
    @mock.patch("PyQt5.QtWidgets.QMessageBox.warning", return_value=QtWidgets.QMessageBox.Yes)
    @mock.patch("PyQt5.QtWidgets.QMessageBox.information", return_value=QtWidgets.QMessageBox.Yes)
    @mock.patch("PyQt5.QtWidgets.QMessageBox.question", return_value=QtWidgets.QMessageBox.Yes)
    @mock.patch("mslib.msui.mss_pyui.get_save_filename", return_value=save_ftml)
    @mock.patch("mslib.msui.mss_pyui.get_open_filename", return_value=save_ftml)
    def test_flight_track_io(self, mockload, mocksave, mockq, mocki, mockw, mockbox):
        self.window.actionCloseSelectedFlightTrack.trigger()
        assert mocki.call_count == 1
        self.window.actionNewFlightTrack.trigger()
        self.window.listFlightTracks.setCurrentRow(0)
        assert self.window.listFlightTracks.count() == 2
        tmp_ft = self.window.active_flight_track
        self.window.active_flight_track = self.window.listFlightTracks.currentItem().flighttrack_model
        self.window.actionCloseSelectedFlightTrack.trigger()
        assert mocki.call_count == 2
        self.window.last_save_directory = self.sample_path
        self.window.actionSaveActiveFlightTrack.trigger()
        self.window.actionSaveActiveFlightTrack.trigger()
        self.window.active_flight_track = tmp_ft
        self.window.actionCloseSelectedFlightTrack.trigger()
        assert self.window.listFlightTracks.count() == 1
        self.window.actionImportFlightTrackFTML.trigger()
        assert self.window.listFlightTracks.count() == 2
        assert os.path.exists(self.save_ftml)
        os.remove(self.save_ftml)



# -*- coding: utf-8 -*-
"""

    mslib.mswms._tests.test_convert
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module provides pytest functions to tests mswms.utils convert_to function.

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2017 Reimar Bauer
    :copyright: Copyright 2016-2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import pytest
from mslib.utils.units import convert_to, units


def test_units():
    assert units("10 degree_north").to("degree").m == 10
    assert units("10 degrees_north").to("degree").m == 10
    assert units("10 degreeN").to("degree").m == 10
    assert units("10 degree_N").to("degree").m == 10
    assert units("-10 degree_south").to("degree").m == 10
    assert units("-10 degrees_south").to("degree").m == 10
    assert units("-10 degreeS").to("degree").m == 10
    assert units("-10 degree_S").to("degree").m == 10
    assert units("10 degree_east").to("degree").m == 10
    assert units("10 degrees_east").to("degree").m == 10
    assert units("10 degreeE").to("degree").m == 10
    assert units("10 degree_E").to("degree").m == 10
    assert units("-10 degree_west").to("degree").m == 10
    assert units("-10 degrees_west").to("degree").m == 10
    assert units("-10 degreeW").to("degree").m == 10
    assert units("-10 degree_W").to("degree").m == 10
    assert units("1 percent"). to("dimensionless").m == 0.01
    assert units("1 permille"). to("dimensionless").m == 0.001
    assert units("1 ppm"). to("dimensionless").m == pytest.approx(1e-6)
    assert units("1 ppb"). to("dimensionless").m == pytest.approx(1e-9)
    assert units("1 ppt"). to("dimensionless").m == pytest.approx(1e-12)
    assert units("1 ppmv"). to("dimensionless").m == pytest.approx(1e-6)
    assert units("1 ppbv"). to("dimensionless").m == pytest.approx(1e-9)
    assert units("1 pptv"). to("dimensionless").m == pytest.approx(1e-12)

    assert units("1 PVU").to_base_units().m == pytest.approx(units("1E-6 m^2 s^-1 K kg^-1").to_base_units().m)


def test_convert_to():
    assert convert_to(10, "km", "m", None) == 10000
    assert convert_to(1000, "m", "km", None) == 1

    assert convert_to(1000, "Pa", "hPa", None) == 10
    assert convert_to(1000, "hPa", "Pa", None) == 100000

    assert convert_to(10, "degC", "K", None) == 283.15

    assert convert_to(10 * 9.81, "m^2s^-2", "m", None) == pytest.approx(10)
    assert convert_to(10 * 9.81, "m**2s**-2", "m", None) == pytest.approx(10)

    with pytest.raises(TypeError):
        assert convert_to(10, "m", "m**2s**-2", None) == 10

    assert convert_to(10, "m", "m**2s**-2", 1) == 10
    assert convert_to(10, "m", "m**2s**-2", 9.81) == pytest.approx(98.1)

    assert convert_to(1000, "", "Pa", 999) == 999000
    assert convert_to(1000, "Pa", "", 999) == 999000
    assert convert_to(1000, "whattheheck", "Pa", 999) == 999000
    assert convert_to(1000, "hPa", "whattheheck", 999) == 999000
    assert convert_to(1000, "whattheheck", "whatthehock", 999) == 999000

    assert convert_to(10, "percent", "dimensionless", None) == 0.1
    assert convert_to(10, "permille", "dimensionless", None) == 0.01
    assert convert_to(10, "ppm", "dimensionless", None) == pytest.approx(10e-6)
    assert convert_to(10, "ppb", "dimensionless", None) == pytest.approx(10e-9)
    assert convert_to(10, "ppt", "dimensionless", None) == pytest.approx(10e-12)
    assert convert_to(10, "ppm", "ppt", None) == pytest.approx(10e6)
    assert convert_to(10, "ppb", "ppm", None) == pytest.approx(10e-3)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mslib/msui/ui/ui_performance_dockwidget.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_PerformanceDockWidget(object):
    def setupUi(self, PerformanceDockWidget):
        PerformanceDockWidget.setObjectName("PerformanceDockWidget")
        PerformanceDockWidget.resize(767, 123)
        self.gridLayout = QtWidgets.QGridLayout(PerformanceDockWidget)
        self.gridLayout.setContentsMargins(9, 9, 9, 9)
        self.gridLayout.setVerticalSpacing(1)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setContentsMargins(-1, 8, -1, -1)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_2 = QtWidgets.QLabel(PerformanceDockWidget)
        self.label_2.setIndent(5)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_2.addWidget(self.label_2)
        self.lbAircraftName = QtWidgets.QLabel(PerformanceDockWidget)
        self.lbAircraftName.setIndent(3)
        self.lbAircraftName.setObjectName("lbAircraftName")
        self.horizontalLayout_2.addWidget(self.lbAircraftName, 0, QtCore.Qt.AlignHCenter)
        self.pbLoadPerformance = QtWidgets.QPushButton(PerformanceDockWidget)
        self.pbLoadPerformance.setObjectName("pbLoadPerformance")
        self.horizontalLayout_2.addWidget(self.pbLoadPerformance, 0, QtCore.Qt.AlignRight)
        self.gridLayout.addLayout(self.horizontalLayout_2, 1, 1, 1, 1)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setContentsMargins(-1, -1, 0, -1)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_4 = QtWidgets.QLabel(PerformanceDockWidget)
        self.label_4.setIndent(1)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_4.addWidget(self.label_4)
        self.dsbEmptyWeight = QtWidgets.QDoubleSpinBox(PerformanceDockWidget)
        self.dsbEmptyWeight.setDecimals(0)
        self.dsbEmptyWeight.setMaximum(9999999.0)
        self.dsbEmptyWeight.setSingleStep(1000.0)
        self.dsbEmptyWeight.setObjectName("dsbEmptyWeight")
        self.horizontalLayout_4.addWidget(self.dsbEmptyWeight)
        self.gridLayout.addLayout(self.horizontalLayout_4, 3, 0, 1, 1)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_3 = QtWidgets.QLabel(PerformanceDockWidget)
        self.label_3.setIndent(1)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_3.addWidget(self.label_3)
        self.dteTakeoffTime = QtWidgets.QDateTimeEdit(PerformanceDockWidget)
        self.dteTakeoffTime.setObjectName("dteTakeoffTime")
        self.horizontalLayout_3.addWidget(self.dteTakeoffTime)
        self.gridLayout.addLayout(self.horizontalLayout_3, 0, 0, 1, 1)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setContentsMargins(-1, 4, -1, -1)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_5 = QtWidgets.QLabel(PerformanceDockWidget)
        self.label_5.setIndent(1)
        self.label_5.setObjectName("label_5")
        self.horizontalLayout_5.addWidget(self.label_5)
        self.dsbTakeoffWeight = QtWidgets.QDoubleSpinBox(PerformanceDockWidget)
        self.dsbTakeoffWeight.setDecimals(0)
        self.dsbTakeoffWeight.setMaximum(9999999.0)
        self.dsbTakeoffWeight.setSingleStep(1000.0)
        self.dsbTakeoffWeight.setObjectName("dsbTakeoffWeight")
        self.horizontalLayout_5.addWidget(self.dsbTakeoffWeight)
        self.gridLayout.addLayout(self.horizontalLayout_5, 1, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.cbShowPerformance = QtWidgets.QCheckBox(PerformanceDockWidget)
        self.cbShowPerformance.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.cbShowPerformance.setTristate(False)
        self.cbShowPerformance.setObjectName("cbShowPerformance")
        self.horizontalLayout.addWidget(self.cbShowPerformance)
        self.gridLayout.addLayout(self.horizontalLayout, 0, 1, 1, 1)

        self.retranslateUi(PerformanceDockWidget)
        QtCore.QMetaObject.connectSlotsByName(PerformanceDockWidget)

    def retranslateUi(self, PerformanceDockWidget):
        _translate = QtCore.QCoreApplication.translate
        PerformanceDockWidget.setWindowTitle(_translate("PerformanceDockWidget", "PerformanceDockWidget"))
        self.label_2.setText(_translate("PerformanceDockWidget", "Aircraft:"))
        self.lbAircraftName.setText(_translate("PerformanceDockWidget", "DUMMY"))
        self.pbLoadPerformance.setText(_translate("PerformanceDockWidget", "Select"))
        self.label_4.setText(_translate("PerformanceDockWidget", "Aircraft weight (no fuel) (lb)"))
        self.label_3.setText(_translate("PerformanceDockWidget", "Take off time"))
        self.label_5.setText(_translate("PerformanceDockWidget", "Maximum take off weight (lb)"))
        self.cbShowPerformance.setText(_translate("PerformanceDockWidget", "Show Performance"))



# -*- coding: utf-8 -*-
"""

    mslib.msui.mpl_pathinteractor
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Interactive editing of Path objects on a Matplotlib canvas.

    This module provides the following classes:

    a) WaypointsPath and subclasses PathV, PathH and PathH-GC:
    Derivatives of matplotlib.Path, provide additional methods to
    insert and delete vertices and to find best insertion points for new
    vertices.

    b) PathInteractor and subclasses VPathInteractor and HPathInteractor:
    Classes that implement a path editor by binding matplotlib mouse events
    to a WaypointsPath object. Support for moving, inserting and deleting vertices.

    The code in this module is inspired by the matplotlib example 'path_editor.py'
    (http://matplotlib.sourceforge.net/examples/event_handling/path_editor.html).

    For more information on implementing animated graphics in matplotlib, see
    http://www.scipy.org/Cookbook/Matplotlib/Animations.


    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import logging
import math
import numpy as np
import matplotlib.path as mpath
import matplotlib.patches as mpatches
from PyQt5 import QtCore, QtWidgets

from mslib.utils.coordinate import get_distance, find_location, latlon_points
from mslib.utils.units import units
from mslib.utils.thermolib import pressure2flightlevel
from mslib.msui import flighttrack as ft


def distance_point_linesegment(p, l1, l2):
    """Computes the distance between a point p and a line segment given by its
       endpoints l1 and l2.

    p, l1, l2 should be numpy arrays of length 2 representing [x,y].

    Based on the dot product formulation from
      'Subject 1.02: How do I find the distance from a point to a line?'
      (from  http://www.faqs.org/faqs/graphics/algorithms-faq/).

    Special case: The point p projects to an extension of the line segment.
    In this case, the distance between the point and the segment equals
    the distance between the point and the closest segment endpoint.
    """
    # Compute the parameter r in the line formulation p* = l1 + r*(l2-l1).
    # p* is the point on the line at which (p-p*) and (l1-l2) form a right
    # angle.
    r = (np.dot(p - l1, l2 - l1) / np.linalg.norm(l2 - l1) ** 2)
    # If 0 < r < 1, we return the distance p-p*. If r > 1, p* is on the
    # forward extension of l1-l2, hence return the distance between
    # p and l2. If r < 0, return the distance p-l1.
    if r > 1:
        return np.linalg.norm(p - l2)
    elif r < 0:
        return np.linalg.norm(p - l1)
    else:
        p_on_line = l1 + r * (l2 - l1)
        return np.linalg.norm(p - p_on_line)


class WaypointsPath(mpath.Path):
    """Derivative of matplotlib.Path that provides methods to insert and
       delete vertices.
    """

    def delete_vertex(self, index):
        """Remove the vertex at the given index from the list of vertices.
        """
        # TODO: Should codes (MOVETO, LINETO) be modified here? relevant for
        #      inserting/deleting first/last point.
        # Emulate pop() for ndarray:
        self.vertices = np.delete(self.vertices, index, axis=0)
        self.codes = np.delete(self.codes, index, axis=0)

    def insert_vertex(self, index, vertex, code):
        """Insert a new vertex (a tuple x,y) with the given code (see
           matplotlib.Path) at the given index.
        """
        self.vertices = np.insert(self.vertices, index,
                                  np.asarray(vertex, np.float_), axis=0)
        self.codes = np.insert(self.codes, index, code, axis=0)

    def index_of_closest_segment(self, x, y, eps=5):
        """Find the index of the edge closest to the specified point at x,y.

        If the point is not within eps (in the same coordinates as x,y) of
        any edge in the path, the index of the closest end point is returned.
        """
        # If only one point is stored in the list the best index to insert a
        # new point is after this point.
        if len(self.vertices) == 1:
            return 1

        # Compute the distance between the first point in the path and
        # the given point.
        point = np.array([x, y])
        min_index = 0
        min_distance = np.linalg.norm(point - self.vertices[0])

        # Loop over all line segments. If the distance between the given
        # point and the segment is smaller than a specified threshold AND
        # the distance is smaller than the currently smallest distance
        # then remember the current index.
        for i in range(len(self.vertices) - 1):
            l1 = self.vertices[i]
            l2 = self.vertices[i + 1]
            distance = distance_point_linesegment(point, l1, l2)
            if distance < eps and distance < min_distance:
                min_index = i + 1
                min_distance = distance

        # Compute the distance between the given point and the end point of
        # the path. Is it smaller than the currently smallest distance?
        distance = np.linalg.norm(point - self.vertices[-1])
        if distance < min_distance:
            min_index = len(self.vertices)

        return min_index

    def transform_waypoint(self, wps_list, index):
        """Transform the waypoint at index <index> of wps_list to plot
           coordinates.

        wps_list is a list of <Waypoint> objects (obtained from
        WaypointsTableModel.allWaypointData()).

        NEEDS TO BE IMPLEMENTED IN DERIVED CLASSES.
        """
        return (0, 0)

    def update_from_WaypointsTableModel(self, wps_model):
        """
        """
        Path = mpath.Path
        wps = wps_model.all_waypoint_data()
        pathdata = []
        # on a expired mscolab server wps is an empty list
        if len(wps) > 0:
            pathdata = [(Path.MOVETO, self.transform_waypoint(wps, 0))]
            for i, _ in enumerate(wps[1:]):
                pathdata.append((Path.LINETO, self.transform_waypoint(wps, i + 1)))

        codes, vertices = list(zip(*pathdata))
        self.codes = np.array(codes, dtype=np.uint8)
        self.vertices = np.array(vertices)


#
# CLASS PathV
#


class PathV(WaypointsPath):
    """Class to represent a vertical flight profile path.
    """

    def __init__(self, *args, **kwargs):
        """The constructor required the additional keyword 'numintpoints':

        numintpoints -- number of intermediate interpolation points. The entire
                        flight track will be interpolated to this number of
                        points.
        """
        self.numintpoints = kwargs.pop("numintpoints")
        super().__init__(*args, **kwargs)

    def update_from_WaypointsTableModel(self, wps_model):
        """Extended version of the corresponding WaypointsPath method.

        The idea is to generate a field 'intermediate_indexes' that stores
        the indexes of the waypoints in the field of the intermediate
        great circle points generated by the flight track model, then
        to use these great circle indexes as x-coordinates for the vertical
        section. This means: If ngc_points are created by
        wps_model.intermediatePoints(), the waypoints are mapped to the
        range 0..ngc_points.

        NOTE: If wps_model only contains two equal waypoints,
        intermediate_indexes will NOT span the entire vertical section
        plot (this is intentional, as a flight with two equal
        waypoints makes no sense).
        """
        # Compute intermediate points.
        lats, lons, times = wps_model.intermediate_points(
            numpoints=self.numintpoints, connection="greatcircle")

        if lats is not None:
            # Determine indices of waypoints in list of intermediate points.
            # Store these indices.
            waypoints = [[wp.lat, wp.lon] for wp in wps_model.all_waypoint_data()]
            intermediate_indexes = []
            ipoint = 0
            for i, (lat, lon) in enumerate(zip(lats, lons)):
                if abs(lat - waypoints[ipoint][0]) < 1E-10 and abs(lon - waypoints[ipoint][1]) < 1E-10:
                    intermediate_indexes.append(i)
                    ipoint += 1
                if ipoint >= len(waypoints):
                    break

            self.intermediate_indexes = intermediate_indexes
            self.ilats = lats
            self.ilons = lons
            self.itimes = times

            # Call super method.
            super().update_from_WaypointsTableModel(wps_model)

    def transform_waypoint(self, wps_list, index):
        """Returns the x-index of the waypoint and its pressure.
        """
        return (self.intermediate_indexes[index], wps_list[index].pressure)


#
# CLASS PathH, PathH_GC
#


class PathH(WaypointsPath):
    """Class to represent a horizontal flight track path, waypoints connected
       by linear line segments.
    """

    def __init__(self, *args, **kwargs):
        """The constructor required the additional keyword 'map' to transform
           vertex cooredinates between lat/lon and projection coordinates.
        """
        self.map = kwargs.pop("map")
        super().__init__(*args, **kwargs)

    def transform_waypoint(self, wps_list, index):
        """Transform lon/lat to projection coordinates.
        """
        return self.map(wps_list[index].lon, wps_list[index].lat)


class PathH_GC(PathH):
    """Class to represent a horizontal flight track path, waypoints connected
       by great circle segments.

    Derives from PathH.

    Provides to kinds of vertex data: (1) Waypoint vertices (wp_vertices) and
    (2) intermediate great circle vertices (vertices).
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.wp_codes = np.array([], dtype=np.uint8)
        self.wp_vertices = np.array([])

    def update_from_WaypointsTableModel(self, wps_model):
        """Get waypoint coordinates from flight track model, get
           intermediate great circle vertices from map instance.
        """
        Path = mpath.Path

        # Waypoint coordinates.
        wps = wps_model.all_waypoint_data()
        if len(wps) > 0:
            pathdata = [(Path.MOVETO, self.transform_waypoint(wps, 0))]
            for i in range(len(wps[1:])):
                pathdata.append((Path.LINETO, self.transform_waypoint(wps, i + 1)))
            wp_codes, wp_vertices = list(zip(*pathdata))
            self.wp_codes = np.array(wp_codes, dtype=np.uint8)
            self.wp_vertices = np.array(wp_vertices)

            # Coordinates of intermediate great circle points.
            lons, lats = list(zip(*[(wp.lon, wp.lat) for wp in wps]))
            x, y = self.map.gcpoints_path(lons, lats)

            if len(x) > 0:
                pathdata = [(Path.MOVETO, (x[0], y[0]))]
                for i in range(len(x[1:])):
                    pathdata.append((Path.LINETO, (x[i + 1], y[i + 1])))
            codes, vertices = list(zip(*pathdata))
            self.codes = np.array(codes, dtype=np.uint8)
            self.vertices = np.array(vertices)

    def index_of_closest_segment(self, x, y, eps=5):
        """Find the index of the edge closest to the specified point at x,y.

        If the point is not within eps (in the same coordinates as x,y) of
        any edge in the path, the index of the closest end point is returned.
        """
        # Determine the index of the great circle vertex that is closest to the
        # given point.
        gcvertex_index = super().index_of_closest_segment(x, y, eps)

        # Determine the waypoint index that corresponds to the great circle
        # index. If the best index is to append the waypoint to the end of
        # the flight track, directly return this index.
        if gcvertex_index == len(self.vertices):
            return len(self.wp_vertices)
        # Otherwise iterate through the list of great circle points and remember
        # the index of the last "real" waypoint that was encountered in this
        # list.
        i = 0  # index for great circle points
        j = 0  # index for waypoints
        wp_vertex = self.wp_vertices[j]
        while (i < gcvertex_index):
            vertex = self.vertices[i]
            if vertex[0] == wp_vertex[0] and vertex[1] == wp_vertex[1]:
                j += 1
                wp_vertex = self.wp_vertices[j]
            i += 1
        return j


#
# CLASS PathInteractor
#


class PathInteractor(QtCore.QObject):
    """An interactive matplotlib path editor. Allows vertices of a path patch
       to be interactively picked and moved around.

    Superclass for the path editors used by the top and side views of the
    Mission Support System.
    """

    showverts = True  # show the vertices of the path patch
    epsilon = 12

    # picking points

    def __init__(self, ax=None, waypoints=None, mplpath=None,
                 facecolor='blue', edgecolor='yellow',
                 linecolor='blue', markerfacecolor='red',
                 marker='o', label_waypoints=True):
        """The constructor initializes the path patches, overlying line
           plot and connects matplotlib signals.

        Arguments:
        ax -- matplotlib.Axes object into which the path should be drawn.
        waypoints -- flighttrack.WaypointsModel instance.
        mplpath -- matplotlib.path.Path instance
        facecolor -- facecolor of the patch
        edgecolor -- edgecolor of the patch
        linecolor -- color of the line plotted above the patch edges
        markerfacecolor -- color of the markers that represent the waypoints
        marker -- symbol of the markers that represent the waypoints, see
                  matplotlib plot() or scatter() routines for more information.
        label_waypoints -- put labels with the waypoint numbers on the waypoints.
        """
        QtCore.QObject.__init__(self)
        self.waypoints_model = None
        self.background = None
        self._ind = None  # the active vertex

        # Create a PathPatch representing the interactively editable path
        # (vertical profile or horizontal flight track in subclasses).
        path = mplpath
        pathpatch = mpatches.PathPatch(path, facecolor=facecolor,
                                       edgecolor=edgecolor, alpha=0.15)
        ax.add_patch(pathpatch)

        self.ax = ax
        self.path = path
        self.pathpatch = pathpatch
        self.pathpatch.set_animated(True)  # ensure correct redrawing

        # Draw the line representing flight track or profile (correct
        # vertices handling for the line needs to be ensured in subclasses).
        x, y = list(zip(*self.pathpatch.get_path().vertices))
        self.line, = self.ax.plot(x, y, color=linecolor,
                                  marker=marker, linewidth=2,
                                  markerfacecolor=markerfacecolor,
                                  animated=True)

        # List to accomodate waypoint labels.
        self.wp_labels = []
        self.label_waypoints = label_waypoints

        # Connect mpl events to handler routines: mouse movements and picks.
        canvas = self.ax.figure.canvas
        canvas.mpl_connect('draw_event', self.draw_callback)
        self.canvas = canvas

        # Set the waypoints model, connect to the change() signals of the model
        # and redraw the figure.
        self.set_waypoints_model(waypoints)

    def set_waypoints_model(self, waypoints):
        """Change the underlying waypoints data structure. Disconnect change()
           signals of an already existing model and connect to the new model.
           Redraw the map.
        """
        # If a model exists, disconnect from the old change() signals.
        wpm = self.waypoints_model
        if wpm:
            wpm.dataChanged.disconnect(self.qt_data_changed_listener)
            wpm.rowsInserted.disconnect(self.qt_insert_remove_point_listener)
            wpm.rowsRemoved.disconnect(self.qt_insert_remove_point_listener)
        # Set the new waypoints model.
        self.waypoints_model = waypoints
        # Connect to the new model's signals.
        wpm = self.waypoints_model
        wpm.dataChanged.connect(self.qt_data_changed_listener)
        wpm.rowsInserted.connect(self.qt_insert_remove_point_listener)
        wpm.rowsRemoved.connect(self.qt_insert_remove_point_listener)
        # Redraw.
        self.pathpatch.get_path().update_from_WaypointsTableModel(wpm)
        self.redraw_figure()

    def qt_insert_remove_point_listener(self, index, first, last):
        """Listens to rowsInserted() and rowsRemoved() signals emitted
           by the flight track data model. The view can thus react to
           data changes induced by another view (table, side view).
        """
        self.pathpatch.get_path().update_from_WaypointsTableModel(self.waypoints_model)
        self.redraw_figure()

    def qt_data_changed_listener(self, index1, index2):
        """Listens to dataChanged() signals emitted by the flight track
           data model. The view can thus react to data changes induced
           by another view (table, top view).
        """
        # REIMPLEMENT IN SUBCLASSES.
        pass

    def draw_callback(self, event):
        """Called when the figure is redrawn. Stores background data (for later
           restoration) and draws artists.
        """
        self.background = self.canvas.copy_from_bbox(self.ax.bbox)
        try:
            # TODO review
            self.ax.draw_artist(self.pathpatch)
        except ValueError as ex:
            # When using Matplotlib 1.2, "ValueError: Invalid codes array."
            # occurs. The error occurs in Matplotlib's backend_agg.py/draw_path()
            # function. However, when I print the codes array in that function,
            # it looks fine -- correct length and correct codes. I can't figure
            # out why that error occurs.. (mr, 2013Feb08).
            logging.error("%s %s", ex, type(ex))
        self.ax.draw_artist(self.line)
        for t in self.wp_labels:
            self.ax.draw_artist(t)
            # The blit() method makes problems (distorted figure background). However,
            # I don't see why it is needed -- everything seems to work without this line.
            # (see infos on http://www.scipy.org/Cookbook/Matplotlib/Animations).
            # self.canvas.blit(self.ax.bbox)

    def get_ind_under_point(self, event):
        """Get the index of the waypoint vertex under the point
           specified by event within epsilon tolerance.

        Uses display coordinates.
        If no waypoint vertex is found, None is returned.
        """
        xy = np.asarray(self.pathpatch.get_path().vertices)
        xyt = self.pathpatch.get_transform().transform(xy)
        xt, yt = xyt[:, 0], xyt[:, 1]
        d = np.hypot(xt - event.x, yt - event.y)
        ind = d.argmin()
        if d[ind] >= self.epsilon:
            ind = None
        return ind

    def button_press_callback(self, event):
        """Called whenever a mouse button is pressed. Determines the index of
           the vertex closest to the click, as long as a vertex is within
           epsilon tolerance of the click.
        """
        if not self.showverts:
            return
        if event.inaxes is None:
            return
        if event.button != 1:
            return
        self._ind = self.get_ind_under_point(event)

    def set_vertices_visible(self, showverts=True):
        """Set the visibility of path vertices (the line plot).
        """
        self.showverts = showverts
        self.line.set_visible(self.showverts)
        for t in self.wp_labels:
            t.set_visible(showverts and self.label_waypoints)
        if not self.showverts:
            self._ind = None
        self.canvas.draw()

    def set_patch_visible(self, showpatch=True):
        """Set the visibility of path patch (the area).
        """
        self.pathpatch.set_visible(showpatch)
        self.canvas.draw()

    def set_labels_visible(self, visible=True):
        """Set the visibility of the waypoint labels.
        """
        self.label_waypoints = visible
        for t in self.wp_labels:
            t.set_visible(self.showverts and self.label_waypoints)
        self.canvas.draw()

    def redraw_path(self, vertices=None):
        """Redraw the matplotlib artists that represent the flight track
           (path patch and line).

        If vertices are specified, they will be applied to the graphics
        output. Otherwise the vertex array obtained from the path patch
        will be used.
        """
        if vertices is None:
            vertices = self.pathpatch.get_path().vertices
        self.line.set_data(list(zip(*vertices)))

        # Draw waypoint labels.
        for wp in self.wp_labels:
            wp.remove()
        self.wp_labels = []  # remove doesn't seem to be necessary
        x, y = list(zip(*vertices))
        wpd = self.waypoints_model.all_waypoint_data()
        for i in range(len(wpd)):
            textlabel = f"{str(i):}   "
            if wpd[i].location != "":
                textlabel = f"{wpd[i].location:}   "
            t = self.ax.text(x[i],
                             y[i],
                             textlabel,
                             bbox=dict(boxstyle="round",
                                       facecolor="white",
                                       alpha=0.5,
                                       edgecolor="none"),
                             fontweight="bold",
                             zorder=4,
                             rotation=90,
                             animated=True,
                             clip_on=True,
                             visible=self.showverts and self.label_waypoints)
            self.wp_labels.append(t)

        if self.background:
            self.canvas.restore_region(self.background)
        try:
            self.ax.draw_artist(self.pathpatch)
        except ValueError as error:
            logging.error("ValueError Exception %s", error)
        self.ax.draw_artist(self.line)
        for t in self.wp_labels:
            self.ax.draw_artist(t)
        self.canvas.blit(self.ax.bbox)

    redraw_figure = redraw_path

    def confirm_delete_waypoint(self, row):
        """Open a QMessageBox and ask the user if he really wants to
           delete the waypoint at index <row>.

        Returns TRUE if the user confirms the deletion.

        If the flight track consists of only two points deleting a waypoint
        is not possible. In this case the user is informed correspondingly.
        """
        wps = self.waypoints_model.all_waypoint_data()
        if len(wps) < 3:
            QtWidgets.QMessageBox.warning(
                None, "Remove waypoint",
                "Cannot remove waypoint, the flight track needs to consist "
                "of at least two points.")
            return False
        else:
            wp = wps[row]
            return QtWidgets.QMessageBox.question(
                None, "Remove waypoint",
                f"Remove waypoint no.{row:d} at {wp.lat:.2f}/{wp.lon:.2f}, flightlevel {wp.flightlevel:.2f}?",
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No) == QtWidgets.QMessageBox.Yes

    def set_path_color(self, line_color=None, marker_facecolor=None,
                       patch_facecolor=None):
        """Set the color of the path patch elements.

        Arguments (options):
        line_color -- color of the path line
        marker_facecolor -- color of the waypoints
        patch_facecolor -- color of the patch covering the path area
        """
        if line_color is not None:
            self.line.set_color(line_color)
        if marker_facecolor is not None:
            self.line.set_markerfacecolor(marker_facecolor)
        if patch_facecolor is not None:
            self.pathpatch.set_facecolor(patch_facecolor)


#
# CLASS VPathInteractor
#


class VPathInteractor(PathInteractor):
    """Subclass of PathInteractor that implements an interactively editable
       vertical profile of the flight track.
    """
    signal_get_vsec = QtCore.Signal(name="get_vsec")

    def __init__(self, ax, waypoints, redraw_xaxis=None, clear_figure=None, numintpoints=101):
        """Constructor passes a PathV instance its parent.

        Arguments:
        ax -- matplotlib.Axes object into which the path should be drawn.
        waypoints -- flighttrack.WaypointsModel instance.
        numintpoints -- number of intermediate interpolation points. The entire
                        flight track will be interpolated to this number of
                        points.
        redrawXAxis -- callback function to redraw the x-axis on path changes.
        """
        self.numintpoints = numintpoints
        self.redraw_xaxis = redraw_xaxis
        self.clear_figure = clear_figure
        super().__init__(
            ax=ax, waypoints=waypoints, mplpath=PathV([[0, 0]], numintpoints=numintpoints))

    def get_num_interpolation_points(self):
        return self.numintpoints

    def redraw_figure(self):
        """For the side view, changes in the horizontal position of a waypoint
           (including moved waypoints, new or deleted waypoints) make a complete
           redraw of the figure necessary.

           Calls the callback function 'redrawXAxis()'.
        """
        self.redraw_path()
        # emit signal to redraw map
        self.signal_get_vsec.emit()
        if self.clear_figure() is not None:
            self.clear_figure()

        if self.redraw_xaxis is not None:
            try:
                self.redraw_xaxis(self.path.ilats, self.path.ilons, self.path.itimes)
            except AttributeError as err:
                logging.debug("%s" % err)
        self.ax.figure.canvas.draw()

    def button_release_delete_callback(self, event):
        """Called whenever a mouse button is released.
        """
        if not self.showverts or event.button != 1:
            return

        if self._ind is not None:
            if self.confirm_delete_waypoint(self._ind):
                # removeRows() will trigger a signal that will redraw the path.
                self.waypoints_model.removeRows(self._ind)
            self._ind = None

    def button_release_insert_callback(self, event):
        """Called whenever a mouse button is released.

        From the click event's coordinates, best_index is calculated as
        the index of a vertex whose x coordinate > clicked x coordinate.
        This is the position where the waypoint is to be inserted.

        'lat' and 'lon' are calculated as an average of each of the first waypoint
        in left and right neighbourhood of inserted waypoint.

        The coordinates are checked against "locations" defined in mss' config.

        A new waypoint with the coordinates, and name is inserted into the waypoints_model.
        """
        if not self.showverts or event.button != 1 or event.inaxes is None:
            return
        y = event.ydata
        wpm = self.waypoints_model
        flightlevel = float(pressure2flightlevel(y * units.Pa).magnitude)
        [lat, lon], best_index = self.get_lat_lon(event)
        loc = find_location(lat, lon)  # skipped tolerance which uses appropriate_epsilon_km
        if loc is not None:
            (lat, lon), location = loc
        else:
            location = ""
        new_wp = ft.Waypoint(lat, lon, flightlevel, location=location)
        wpm.insertRows(best_index, rows=1, waypoints=[new_wp])
        self.redraw_figure()

        self._ind = None

    def get_lat_lon(self, event):
        x = event.xdata
        wpm = self.waypoints_model
        vertices = self.pathpatch.get_path().vertices
        best_index = 1
        # if x axis has increasing coordinates
        if vertices[-1, 0] > vertices[0, 0]:
            for index, vertex in enumerate(vertices):
                if x >= vertex[0]:
                    best_index = index + 1
        # if x axis has decreasing coordinates
        else:
            for index, vertex in enumerate(vertices):
                if x <= vertex[0]:
                    best_index = index + 1
        # number of subcoordinates is determined by difference in x coordinates
        number_of_intermediate_points = math.floor(vertices[best_index, 0] - vertices[best_index - 1, 0])
        vert_xs, vert_ys = latlon_points(
            vertices[best_index - 1, 0], vertices[best_index - 1, 1],
            vertices[best_index, 0], vertices[best_index, 1],
            number_of_intermediate_points, connection="linear")
        lats, lons = latlon_points(
            wpm.waypoint_data(best_index - 1).lat, wpm.waypoint_data(best_index - 1).lon,
            wpm.waypoint_data(best_index).lat, wpm.waypoint_data(best_index).lon,
            number_of_intermediate_points, connection="greatcircle")

        # best_index1 is the best index among the intermediate coordinates to fit the hovered point
        # if x axis has increasing coordinates
        best_index1 = np.argmin(abs(vert_xs - x))
        # depends if best_index1 or best_index1 - 1 on closeness to left or right neighbourhood
        return (lats[best_index1], lons[best_index1]), best_index

    def button_release_move_callback(self, event):
        """Called whenever a mouse button is released.
        """
        if not self.showverts or event.button != 1:
            return

        if self._ind is not None:
            # Submit the new pressure (the only value that can be edited
            # in the side view) to the data model.
            vertices = self.pathpatch.get_path().vertices
            pressure = vertices[self._ind, 1]
            # http://doc.trolltech.com/4.3/qabstractitemmodel.html#createIndex
            qt_index = self.waypoints_model.createIndex(self._ind, ft.PRESSURE)
            # NOTE: QVariant cannot handle numpy.float64 types, hence convert
            # to float().
            self.waypoints_model.setData(qt_index, QtCore.QVariant(float(pressure / 100.)))

        self._ind = None

    def motion_notify_callback(self, event):
        """Called on mouse movement. Redraws the path if a vertex has been
           picked and is being dragged.

        In the side view, the horizontal position of a waypoint is locked.
        Hence, points can only be moved in the vertical direction (y position
        in this view).
        """
        if not self.showverts or self._ind is None or event.inaxes is None or event.button != 1:
            return
        vertices = self.pathpatch.get_path().vertices
        # Set the new y position of the vertex to event.ydata. Keep the
        # x coordinate.
        vertices[self._ind] = vertices[self._ind, 0], event.ydata
        self.redraw_path(vertices)

    def qt_data_changed_listener(self, index1, index2):
        """Listens to dataChanged() signals emitted by the flight track
           data model. The side view can thus react to data changes
           induced by another view (table, top view).
        """
        # If the altitude of a point has changed, only the plotted flight
        # profile needs to be redrawn (redraw_path()). If the horizontal
        # position of a waypoint has changed, the entire figure needs to be
        # redrawn, as this affects the x-position of all points.
        self.pathpatch.get_path().update_from_WaypointsTableModel(self.waypoints_model)
        if index1.column() in [ft.FLIGHTLEVEL, ft.PRESSURE, ft.LOCATION]:
            self.redraw_path(self.pathpatch.get_path().vertices)
        elif index1.column() in [ft.LAT, ft.LON]:
            self.redraw_figure()
        elif index1.column() in [ft.TIME_UTC]:
            if self.redraw_xaxis is not None:
                self.redraw_xaxis(self.path.ilats, self.path.ilons, self.path.itimes)


class LPathInteractor(PathInteractor):
    """
    Subclass of PathInteractor that implements a non interactive linear profile of the flight track.
    """
    signal_get_lsec = QtCore.Signal(name="get_lsec")

    def __init__(self, ax, waypoints, redraw_xaxis=None, clear_figure=None, numintpoints=101):
        """Constructor passes a PathV instance its parent.

        Arguments:
        ax -- matplotlib.Axes object into which the path should be drawn.
        waypoints -- flighttrack.WaypointsModel instance.
        numintpoints -- number of intermediate interpolation points. The entire
                        flight track will be interpolated to this number of
                        points.
        redrawXAxis -- callback function to redraw the x-axis on path changes.
        """
        self.numintpoints = numintpoints
        self.clear_figure = clear_figure
        self.redraw_xaxis = redraw_xaxis
        super().__init__(
            ax=ax, waypoints=waypoints, mplpath=PathV([[0, 0]], numintpoints=numintpoints))

    def get_num_interpolation_points(self):
        return self.numintpoints

    def redraw_figure(self):
        """For the linear view, changes in the horizontal or vertical position of a waypoint
           (including moved waypoints, new or deleted waypoints) make a complete
           redraw of the figure necessary.
        """
        # emit signal to redraw map
        self.redraw_xaxis()
        self.signal_get_lsec.emit()

    def redraw_path(self, vertices=None):
        """Skip redrawing paths for LSec
        """
        pass

    def draw_callback(self, event):
        """Skip drawing paths for LSec
        """
        pass

    def get_lat_lon(self, event):
        x = event.xdata
        wpm = self.waypoints_model
        vertices = self.pathpatch.get_path().vertices
        best_index = 1
        # if x axis has increasing coordinates
        if vertices[-1, 0] > vertices[0, 0]:
            for index, vertex in enumerate(vertices):
                if x >= vertex[0]:
                    best_index = index + 1
        # if x axis has decreasing coordinates
        else:
            for index, vertex in enumerate(vertices):
                if x <= vertex[0]:
                    best_index = index + 1
        # number of subcoordinates is determined by difference in x coordinates
        number_of_intermediate_points = int(abs(vertices[best_index, 0] - vertices[best_index - 1, 0]))
        vert_xs, vert_ys = latlon_points(
            vertices[best_index - 1, 0], vertices[best_index - 1, 1],
            vertices[best_index, 0], vertices[best_index, 1],
            number_of_intermediate_points, connection="linear")
        lats, lons = latlon_points(
            wpm.waypoint_data(best_index - 1).lat, wpm.waypoint_data(best_index - 1).lon,
            wpm.waypoint_data(best_index).lat, wpm.waypoint_data(best_index).lon,
            number_of_intermediate_points, connection="greatcircle")
        alts = np.linspace(wpm.waypoint_data(best_index - 1).flightlevel,
                           wpm.waypoint_data(best_index).flightlevel, number_of_intermediate_points)

        best_index1 = np.argmin(abs(vert_xs - x))
        # depends if best_index1 or best_index1 - 1 on closeness to left or right neighbourhood
        return (lats[best_index1], lons[best_index1], alts[best_index1]), best_index

    def qt_data_changed_listener(self, index1, index2):
        """Listens to dataChanged() signals emitted by the flight track
           data model. The linear view can thus react to data changes
           induced by another view (table, top view, side view).
        """
        self.pathpatch.get_path().update_from_WaypointsTableModel(self.waypoints_model)
        self.redraw_figure()

#
# CLASS HPathInteractor
#


class HPathInteractor(PathInteractor):
    """Subclass of PathInteractor that implements an interactively editable
       horizontal flight track. Waypoints are connected with great circles.
    """

    def __init__(self, mplmap, waypoints,
                 linecolor='blue', markerfacecolor='red', show_marker=True,
                 label_waypoints=True):
        """Constructor passes a PathH_GC instance its parent (horizontal path
           with waypoints connected with great circles).

        Arguments:
        mplmap -- mpl_map.MapCanvas instance into which the path should be drawn.
        waypoints -- flighttrack.WaypointsModel instance.
        """
        self.map = mplmap
        self.wp_scatter = None
        self.markerfacecolor = markerfacecolor
        self.tangent_lines = None
        self.show_tangent_points = False
        self.solar_lines = None
        self.show_marker = show_marker
        self.show_solar_angle = None
        self.remote_sensing = None
        super().__init__(
            ax=mplmap.ax, waypoints=waypoints,
            mplpath=PathH_GC([[0, 0]], map=mplmap),
            facecolor='none', edgecolor='none', linecolor=linecolor,
            markerfacecolor=markerfacecolor, marker='',
            label_waypoints=label_waypoints)

    def appropriate_epsilon(self, px=5):
        """Determine an epsilon value appropriate for the current projection and
           figure size.

        The epsilon value gives the distance required in map projection
        coordinates that corresponds to approximately px Pixels in screen
        coordinates. The value can be used to find the line/point that is
        closest to a click while discarding clicks that are too far away
        from any geometry feature.
        """
        # (bounds = left, bottom, width, height)
        ax_bounds = self.ax.bbox.bounds
        width = int(round(ax_bounds[2]))
        map_delta_x = np.hypot(self.map.llcrnry - self.map.urcrnry, self.map.llcrnrx - self.map.urcrnrx)
        map_coords_per_px_x = map_delta_x / width
        return map_coords_per_px_x * px

    def appropriate_epsilon_km(self, px=5):
        """Determine an epsilon value appropriate for the current projection and
           figure size.

        The epsilon value gives the distance required in map projection
        coordinates that corresponds to approximately px Pixels in screen
        coordinates. The value can be used to find the line/point that is
        closest to a click while discarding clicks that are too far away
        from any geometry feature.
        """
        # (bounds = left, bottom, width, height)
        ax_bounds = self.ax.bbox.bounds
        diagonal = math.hypot(round(ax_bounds[2]), round(ax_bounds[3]))
        map_delta = get_distance(self.map.llcrnrlat, self.map.llcrnrlon, self.map.urcrnrlat, self.map.urcrnrlon)
        km_per_px = map_delta / diagonal

        return km_per_px * px

    def get_lat_lon(self, event):
        return self.map(event.xdata, event.ydata, inverse=True)[::-1]

    def button_release_insert_callback(self, event):
        """Called whenever a mouse button is released.

        From the click event's coordinates, best_index is calculated if it can be optimally fit
        as a prior waypoint in the path.

        A vertex with same coordinates is inserted into the path in canvas.

        The coordinates are checked against "locations" defined in mss' config.

        A new waypoint with the coordinates, and name is inserted into the waypoints_model.
        """
        if not self.showverts or event.button != 1 or event.inaxes is None:
            return

        # Get position for new vertex.
        x, y = event.xdata, event.ydata
        best_index = self.pathpatch.get_path().index_of_closest_segment(
            x, y, eps=self.appropriate_epsilon())
        logging.debug("TopView insert point: clicked at (%f, %f), "
                      "best index: %d", x, y, best_index)
        self.pathpatch.get_path().insert_vertex(best_index, [x, y], WaypointsPath.LINETO)

        lon, lat = self.map(x, y, inverse=True)
        loc = find_location(lat, lon, tolerance=self.appropriate_epsilon_km(px=15))
        if loc is not None:
            (lat, lon), location = loc
        else:
            location = ""
        wpm = self.waypoints_model
        if len(wpm.all_waypoint_data()) > 0 and 0 < best_index <= len(wpm.all_waypoint_data()):
            flightlevel = wpm.waypoint_data(best_index - 1).flightlevel
        elif len(wpm.all_waypoint_data()) > 0 and best_index == 0:
            flightlevel = wpm.waypoint_data(0).flightlevel
        else:
            logging.error("Cannot copy flightlevel. best_index: %s, len: %s",
                          best_index, len(wpm.all_waypoint_data()))
            flightlevel = 0
        new_wp = ft.Waypoint(lat, lon, flightlevel, location=location)
        wpm.insertRows(best_index, rows=1, waypoints=[new_wp])
        self.redraw_path()

        self._ind = None

    def button_release_move_callback(self, event):
        """Called whenever a mouse button is released.
        """
        if not self.showverts or event.button != 1 or self._ind is None:
            return

        # Submit the new position to the data model.
        vertices = self.pathpatch.get_path().wp_vertices
        lon, lat = self.map(vertices[self._ind][0], vertices[self._ind][1],
                            inverse=True)
        loc = find_location(lat, lon, tolerance=self.appropriate_epsilon_km(px=15))
        if loc is not None:
            lat, lon = loc[0]
        self.waypoints_model.setData(
            self.waypoints_model.createIndex(self._ind, ft.LAT), QtCore.QVariant(lat), update=False)
        self.waypoints_model.setData(
            self.waypoints_model.createIndex(self._ind, ft.LON), QtCore.QVariant(lon))

        self._ind = None

    def button_release_delete_callback(self, event):
        """Called whenever a mouse button is released.
        """
        if not self.showverts or event.button != 1:
            return

        if self._ind is not None and self.confirm_delete_waypoint(self._ind):
            # removeRows() will trigger a signal that will redraw the path.
            self.waypoints_model.removeRows(self._ind)

        self._ind = None

    def motion_notify_callback(self, event):
        """Called on mouse movement. Redraws the path if a vertex has been
           picked and dragged.
        """
        if not self.showverts:
            return
        if self._ind is None:
            return
        if event.inaxes is None:
            return
        if event.button != 1:
            return
        wp_vertices = self.pathpatch.get_path().wp_vertices
        wp_vertices[self._ind] = event.xdata, event.ydata
        self.redraw_path(wp_vertices)

    def qt_data_changed_listener(self, index1, index2):
        """Listens to dataChanged() signals emitted by the flight track
           data model. The top view can thus react to data changes
           induced by another view (table, side view).
        """
        # Update the top view if the horizontal position of any point has been
        # changed.
        if index1.column() in [ft.LOCATION, ft.LAT, ft.LON, ft.FLIGHTLEVEL]:
            self.update()

    def update(self):
        """Update the path plot by updating coordinates and intermediate
           great circle points from the path patch, then redrawing.
        """
        self.pathpatch.get_path().update_from_WaypointsTableModel(
            self.waypoints_model)
        self.redraw_path()

    def redraw_path(self, wp_vertices=None):
        """Redraw the matplotlib artists that represent the flight track
           (path patch, line and waypoint scatter).

        If waypoint vertices are specified, they will be applied to the
        graphics output. Otherwise the vertex array obtained from the path
        patch will be used.
        """

        if wp_vertices is None:
            wp_vertices = self.pathpatch.get_path().wp_vertices
            if len(wp_vertices) == 0:
                raise IOError("mscolab session expired")
            vertices = self.pathpatch.get_path().vertices
        else:
            # If waypoints have been provided, compute the intermediate
            # great circle points for the line instance.
            x, y = list(zip(*wp_vertices))
            lons, lats = self.map(x, y, inverse=True)
            x, y = self.map.gcpoints_path(lons, lats)
            vertices = list(zip(x, y))

        # Set the line to disply great circle points, remove existing
        # waypoints scatter instance and draw a new one. This is
        # necessary as scatter() does not provide a set_data method.
        self.line.set_data(list(zip(*vertices)))

        wp_heights = [(wp.flightlevel * 0.03048) for wp in self.waypoints_model.all_waypoint_data()]
        wp_times = [wp.utc_time for wp in self.waypoints_model.all_waypoint_data()]

        if self.tangent_lines is not None:
            self.tangent_lines.remove()
        if self.show_tangent_points:
            assert self.remote_sensing is not None
            self.tangent_lines = self.remote_sensing.compute_tangent_lines(
                self.map, wp_vertices, wp_heights)
            self.ax.add_collection(self.tangent_lines)
        else:
            self.tangent_lines = None

        if self.solar_lines is not None:
            self.solar_lines.remove()
        if self.show_solar_angle is not None:
            assert self.remote_sensing is not None
            self.solar_lines = self.remote_sensing.compute_solar_lines(
                self.map, wp_vertices, wp_heights, wp_times, self.show_solar_angle)
            self.ax.add_collection(self.solar_lines)
        else:
            self.solar_lines = None

        if self.wp_scatter is not None:
            self.wp_scatter.remove()
        x, y = list(zip(*wp_vertices))

        if self.map.projection == "cyl":  # hack for wraparound
            x = np.array(x)
            x[x < self.map.llcrnrlon] += 360
            x[x > self.map.urcrnrlon] -= 360
        # (animated is important to remove the old scatter points from the map)
        self.wp_scatter = self.ax.scatter(
            x, y, color=self.markerfacecolor, s=20, zorder=3, animated=True, visible=self.show_marker)

        # Draw waypoint labels.
        label_offset = self.appropriate_epsilon(px=5)
        for wp_label in self.wp_labels:
            wp_label.remove()
        self.wp_labels = []  # remove doesn't seem to be necessary
        wpd = self.waypoints_model.all_waypoint_data()
        for i in range(len(wpd)):
            textlabel = str(i)
            if wpd[i].location != "":
                textlabel = f"{wpd[i].location:}"
            t = self.ax.text(
                x[i] + label_offset, y[i] + label_offset, textlabel,
                bbox={"boxstyle": "round", "facecolor": "white", "alpha": 0.6, "edgecolor": "none"},
                fontweight="bold", zorder=4, animated=True, clip_on=True,
                visible=self.showverts and self.label_waypoints)
            self.wp_labels.append(t)

        # Redraw the artists.
        if self.background:
            self.canvas.restore_region(self.background)
        try:
            self.ax.draw_artist(self.pathpatch)
        except ValueError as error:
            logging.debug("ValueError Exception '%s'", error)
        self.ax.draw_artist(self.line)
        self.ax.draw_artist(self.wp_scatter)

        for t in self.wp_labels:
            self.ax.draw_artist(t)
        if self.show_tangent_points:
            self.ax.draw_artist(self.tangent_lines)
        if self.show_solar_angle is not None:
            self.ax.draw_artist(self.solar_lines)
        self.canvas.blit(self.ax.bbox)

    # Link redraw_figure() to redraw_path().
    redraw_figure = redraw_path

    def draw_callback(self, event):
        """Extends PathInteractor.draw_callback() by drawing the scatter
           instance.
        """
        PathInteractor.draw_callback(self, event)
        self.ax.draw_artist(self.wp_scatter)
        if self.show_solar_angle:
            self.ax.draw_artist(self.solar_lines)
        if self.show_tangent_points:
            self.ax.draw_artist(self.tangent_lines)

    def get_ind_under_point(self, event):
        """Get the index of the waypoint vertex under the point
           specified by event within epsilon tolerance.

        Uses display coordinates.
        If no waypoint vertex is found, None is returned.
        """
        xy = np.asarray(self.pathpatch.get_path().wp_vertices)
        if self.map.projection == "cyl":  # hack for wraparound
            lon_min, lon_max = self.map.llcrnrlon, self.map.urcrnrlon
            xy[xy[:, 0] < lon_min, 0] += 360
            xy[xy[:, 0] > lon_max, 0] -= 360
        xyt = self.pathpatch.get_transform().transform(xy)
        xt, yt = xyt[:, 0], xyt[:, 1]
        d = np.hypot(xt - event.x, yt - event.y)
        ind = d.argmin()
        if d[ind] >= self.epsilon:
            ind = None
        return ind

    def set_path_color(self, line_color=None, marker_facecolor=None,
                       patch_facecolor=None):
        """Set the color of the path patch elements.

        Arguments (options):
        line_color -- color of the path line
        marker_facecolor -- color of the waypoints
        patch_facecolor -- color of the patch covering the path area
        """
        PathInteractor.set_path_color(self, line_color, marker_facecolor,
                                      patch_facecolor)
        if marker_facecolor is not None:
            self.wp_scatter.set_facecolor(marker_facecolor)
            self.wp_scatter.set_edgecolor(marker_facecolor)
            self.markerfacecolor = marker_facecolor

    def set_vertices_visible(self, showverts=True):
        """Set the visibility of path vertices (the line plot).
        """
        self.wp_scatter.set_visible(self.show_marker)
        PathInteractor.set_vertices_visible(self, showverts)

    def set_tangent_visible(self, visible):
        self.show_tangent_points = visible

    def set_solar_angle_visible(self, visible):
        self.show_solar_angle = visible

    def set_remote_sensing(self, ref):
        self.remote_sensing = ref



# -*- coding: utf-8 -*-
"""

    mslib.mscolab._tests.test_seed
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    tests for conf functionalities

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from flask_testing import TestCase

from mslib.mscolab.conf import mscolab_settings
from mslib.mscolab.models import db, User, Operation
from mslib.mscolab.mscolab import handle_db_reset
from mslib.mscolab.server import APP
from mslib.mscolab.file_manager import FileManager
from mslib.mscolab.seed import (add_user, get_user, add_operation, add_user_to_operation,
                                delete_user, delete_operation, add_all_users_default_operation)


class Test_Seed(TestCase):
    render_templates = False

    def create_app(self):
        app = APP
        app.config['SQLALCHEMY_DATABASE_URI'] = mscolab_settings.SQLALCHEMY_DB_URI
        app.config['MSCOLAB_DATA_DIR'] = mscolab_settings.MSCOLAB_DATA_DIR
        app.config['UPLOAD_FOLDER'] = mscolab_settings.UPLOAD_FOLDER
        app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
        app.config["TESTING"] = True
        app.config['LIVESERVER_TIMEOUT'] = 10
        app.config['LIVESERVER_PORT'] = 0
        return app

    def setUp(self):
        handle_db_reset()
        db.init_app(self.app)
        self.fm = FileManager(self.app.config["MSCOLAB_DATA_DIR"])
        self.operation_name = "XYZ"
        self.description = "Template"
        self.userdata_0 = 'UV0@uv0', 'UV0', 'uv0'
        self.userdata_1 = "UV1@uv1", "UV1", "UV1"
        self.userdata_2 = "UV2@v2", "V2", "v2"

        assert add_user(self.userdata_0[0], self.userdata_0[1], self.userdata_0[2])
        assert add_operation(self.operation_name, self.description)
        assert add_user_to_operation(path=self.operation_name, emailid=self.userdata_0[0])
        self.user = User(self.userdata_0[0], self.userdata_0[1], self.userdata_0[2])

    def tearDown(self):
        pass

    def test_add_operation(self):
        with self.app.test_client():
            assert add_operation("a1", "description")
            operation = Operation.query.filter_by(path="a1").first()
            assert operation.id > 0

    def test_delete_operation(self):
        with self.app.test_client():
            assert add_operation("todelete", "description")
            operation = Operation.query.filter_by(path="todelete").first()
            assert operation.id > 0
            assert delete_operation("todelete")
            operation = Operation.query.filter_by(path="todelete").first()
            assert operation is None

    def test_add_all_users_default_operation_viewer(self):
        with self.app.test_client():
            assert add_user(self.userdata_1[0], self.userdata_1[1], self.userdata_1[2])
            # viewer
            add_all_users_default_operation(path='XYZ', description="Operation to keep all users",
                                            access_level='viewer')
            expected_result = [{'access_level': 'viewer', 'category': 'default',
                                'description': 'Template', 'op_id': 7, 'path': 'XYZ'}]
            user = User.query.filter_by(emailid=self.userdata_1[0]).first()
            assert user is not None
            result = self.fm.list_operations(user)
            # we don't care here for op_id
            expected_result[0]['op_id'] = result[0]['op_id']
            assert result == expected_result

    def test_add_all_users_default_operation_collaborator(self):
        with self.app.test_client():
            # collaborator
            assert add_user(self.userdata_1[0], self.userdata_1[1], self.userdata_1[2])
            add_all_users_default_operation(path='XYZ', description="Operation to keep all users",
                                            access_level='collaborator')
            expected_result = [{'access_level': 'collaborator', 'category': 'default',
                                'description': 'Template', 'op_id': 7, 'path': 'XYZ'}]
            user = User.query.filter_by(emailid=self.userdata_1[0]).first()
            assert user is not None
            result = self.fm.list_operations(user)
            # we don't care here for op_id
            expected_result[0]['op_id'] = result[0]['op_id']
            assert result == expected_result

    def test_add_all_users_default_operation_creator(self):
        with self.app.test_client():
            assert add_user(self.userdata_1[0], self.userdata_1[1], self.userdata_1[2])
            # creator
            add_all_users_default_operation(path='XYZ', description="Operation to keep all users",
                                            access_level='creator')
            expected_result = [{'access_level': 'creator', 'category': 'default',
                                'description': 'Template', 'op_id': 7, 'path': 'XYZ'}]
            user = User.query.filter_by(emailid=self.userdata_1[0]).first()
            result = self.fm.list_operations(user)
            # we don't care here for op_id
            expected_result[0]['op_id'] = result[0]['op_id']
            assert result == expected_result

    def test_add_all_users_default_operation_creator_unknown_operation(self):
        with self.app.test_client():
            assert add_user(self.userdata_1[0], self.userdata_1[1], self.userdata_1[2])
            # creator added to new operation
            add_all_users_default_operation(path='UVXYZ', description="Operation to keep all users",
                                            access_level='creator')
            expected_result = [{'access_level': 'creator', 'category': 'default',
                                'description': 'Operation to keep all users',
                                'op_id': 7, 'path': 'UVXYZ'}]
            user = User.query.filter_by(emailid=self.userdata_1[0]).first()
            result = self.fm.list_operations(user)
            # we don't care here for op_id
            expected_result[0]['op_id'] = result[0]['op_id']
            assert result == expected_result

    def test_add_user(self):
        with self.app.test_client():
            assert add_user(self.userdata_2[0], self.userdata_2[1], self.userdata_2[2])
            assert add_user(self.userdata_2[0], self.userdata_2[1], self.userdata_2[2]) is False

    def test_get_user(self):
        with self.app.test_client():
            assert add_user(self.userdata_2[0], self.userdata_2[1], self.userdata_2[2])
            user = get_user(self.userdata_2[0])
            assert user.id is not None
            assert user.emailid == self.userdata_2[0]

    def test_add_user_to_operation(self):
        with self.app.test_client():
            assert add_user(self.userdata_2[0], self.userdata_2[1], self.userdata_2[2])
            assert add_operation("operation2", "description")
            assert add_user_to_operation(path="operation2", access_level='admin', emailid=self.userdata_2[0])

    def test_delete_user(self,):
        with self.app.test_client():
            assert add_user(self.userdata_2[0], self.userdata_2[1], self.userdata_2[2])
            user = User.query.filter_by(emailid=self.userdata_2[0]).first()
            assert user is not None
            assert delete_user(self.userdata_2[0])
            user = User.query.filter_by(emailid=self.userdata_2[0]).first()
            assert user is None



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui_linearview_window.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_LinearWindow(object):
    def setupUi(self, LinearWindow):
        LinearWindow.setObjectName("LinearWindow")
        LinearWindow.resize(800, 600)
        self.centralwidget = QtWidgets.QWidget(LinearWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.mpl = MplLinearViewWidget(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.mpl.sizePolicy().hasHeightForWidth())
        self.mpl.setSizePolicy(sizePolicy)
        self.mpl.setMinimumSize(QtCore.QSize(100, 100))
        self.mpl.setObjectName("mpl")
        self.verticalLayout.addWidget(self.mpl)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.cbTools = QtWidgets.QComboBox(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cbTools.sizePolicy().hasHeightForWidth())
        self.cbTools.setSizePolicy(sizePolicy)
        self.cbTools.setObjectName("cbTools")
        self.cbTools.addItem("")
        self.cbTools.addItem("")
        self.horizontalLayout.addWidget(self.cbTools)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.lvoptionbtn = QtWidgets.QPushButton(self.centralwidget)
        self.lvoptionbtn.setObjectName("lvoptionbtn")
        self.horizontalLayout.addWidget(self.lvoptionbtn)
        self.verticalLayout.addLayout(self.horizontalLayout)
        LinearWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(LinearWindow)
        QtCore.QMetaObject.connectSlotsByName(LinearWindow)

    def retranslateUi(self, LinearWindow):
        _translate = QtCore.QCoreApplication.translate
        LinearWindow.setWindowTitle(_translate("LinearWindow", "Linear View - Mission Support System"))
        self.cbTools.setItemText(0, _translate("LinearWindow", "(select to open control)"))
        self.cbTools.setItemText(1, _translate("LinearWindow", "WMS"))
        self.lvoptionbtn.setText(_translate("LinearWindow", "options"))
from mslib.msui.mpl_qtwidget import MplLinearViewWidget


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    LinearWindow = QtWidgets.QMainWindow()
    ui = Ui_LinearWindow()
    ui.setupUi(LinearWindow)
    LinearWindow.show()
    sys.exit(app.exec_())
